# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import unittest

import numpy as np
import pandas as pd
import xarray as xr

from xcube_resampling.temporal import (
    _get_agg_method_kwargs,
    _get_temporal_agg_method,
    _get_temporal_interp_method,
    _guess_resampling_operation,
    resample_in_time,
)

from .sampledata import create_nx8x6_dataset_with_regular_coords


class ResampleInTimeTest(unittest.TestCase):
    def setUp(self):
        self.regular_time = pd.date_range("2020-01-01", periods=5, freq="D")
        self.irregular_time = pd.to_datetime(
            ["2020-01-01", "2020-01-02", "2020-01-04", "2020-01-07"]
        )
        self.ds_regular = xr.Dataset(
            {"a": ("time", np.arange(5))}, coords={"time": self.regular_time}
        )
        self.ds_irregular = xr.Dataset(
            {
                "a": ("time", np.arange(4)),
                "b": (("y", "x"), np.arange(20).reshape((5, 4))),
            },
            coords={"time": self.irregular_time, "x": np.arange(4), "y": np.arange(5)},
        )
        self.ds_notime = xr.Dataset(
            {"a": (("y", "x"), np.arange(20).reshape((5, 4)))},
            coords={"x": np.arange(4), "y": np.arange(5)},
        )

    def test_resample_in_time_min_max(self):
        input_cube = create_nx8x6_dataset_with_regular_coords(8)
        resampled_cube = resample_in_time(input_cube, "2D", agg_methods=["min", "max"])
        self.assertIn("time", resampled_cube)
        self.assertIn("refl_min", resampled_cube)
        self.assertIn("refl_max", resampled_cube)
        self.assertEqual((4,), resampled_cube.time.shape)
        self.assertEqual(("time", "lat", "lon"), resampled_cube.refl_min.dims)
        self.assertEqual(("time", "lat", "lon"), resampled_cube.refl_max.dims)
        self.assertEqual((4, 6, 8), resampled_cube.refl_min.shape)
        self.assertEqual((4, 6, 8), resampled_cube.refl_max.shape)
        self.assertEqual(
            list(resampled_cube.time.values),
            [
                np.datetime64("2025-08-01T00:00:00.000000000"),
                np.datetime64("2025-08-03T00:00:00.000000000"),
                np.datetime64("2025-08-05T00:00:00.000000000"),
                np.datetime64("2025-08-07T00:00:00.000000000"),
            ],
        )
        np.testing.assert_allclose(
            resampled_cube.refl_min.values[..., 0, 1],
            np.array([-3.0, -1.0, 1.0, 3.0]),
        )
        np.testing.assert_allclose(
            resampled_cube.refl_max.values[..., 0, 1],
            np.array([-2.0, 0.0, 2.0, 4.0]),
        )

    def test_resample_in_time_p90(self):
        input_cube = create_nx8x6_dataset_with_regular_coords(8)
        resampled_cube = resample_in_time(input_cube, "3D", agg_methods="percentile_90")
        self.assertIn("time", resampled_cube)
        self.assertIn("refl", resampled_cube)
        self.assertEqual((3,), resampled_cube.time.shape)
        self.assertEqual(("time", "lat", "lon"), resampled_cube.refl.dims)
        self.assertEqual((3, 6, 8), resampled_cube.refl.shape)
        self.assertEqual(
            list(resampled_cube.time.values),
            [
                np.datetime64("2025-08-01T00:00:00.000000000"),
                np.datetime64("2025-08-04T00:00:00.000000000"),
                np.datetime64("2025-08-07T00:00:00.000000000"),
            ],
        )
        np.testing.assert_allclose(
            resampled_cube.refl.values[..., 0, 1],
            np.array([-1.2, 1.8, 3.9]),
        )

    def test_resample_in_time_all(self):
        input_cube = create_nx8x6_dataset_with_regular_coords(8)
        resampled_cube = resample_in_time(input_cube, "all", agg_methods=["min", "max"])
        self.assertIn("time", resampled_cube)
        self.assertIn("refl_min", resampled_cube)
        self.assertIn("refl_max", resampled_cube)
        self.assertIn("ndvi_min", resampled_cube)
        self.assertIn("ndvi_max", resampled_cube)
        self.assertEqual((1,), resampled_cube.time.shape)
        self.assertEqual(("time", "lat", "lon"), resampled_cube.refl_min.dims)
        self.assertEqual(("time", "lat", "lon"), resampled_cube.refl_max.dims)
        self.assertEqual(("time", "lat", "lon"), resampled_cube.ndvi_min.dims)
        self.assertEqual(("time", "lat", "lon"), resampled_cube.ndvi_max.dims)
        self.assertEqual((1, 6, 8), resampled_cube.refl_min.shape)
        self.assertEqual((1, 6, 8), resampled_cube.refl_max.shape)
        self.assertEqual((1, 6, 8), resampled_cube.ndvi_min.shape)
        self.assertEqual((1, 6, 8), resampled_cube.ndvi_max.shape)
        self.assertEqual(
            list(resampled_cube.time.values),
            [
                np.datetime64("2025-08-01T00:00:00.000000000"),
            ],
        )
        np.testing.assert_allclose(
            resampled_cube.refl_min.values[..., 0, 1],
            np.array([-3.0]),
        )
        np.testing.assert_allclose(
            resampled_cube.refl_max.values[..., 0, 1],
            np.array([4.0]),
        )

    def test_resample_in_time_nearest_interpolation(self):
        input_cube = create_nx8x6_dataset_with_regular_coords(4)
        resampled_cube = resample_in_time(input_cube, "6h", interp_methods="nearest")
        self.assertIn("time", resampled_cube)
        self.assertIn("refl", resampled_cube)
        self.assertIn("ndvi", resampled_cube)
        self.assertEqual((13,), resampled_cube.time.shape)
        np.testing.assert_allclose(
            resampled_cube.refl.values[..., 0, 1],
            np.array([-1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2]),
        )

    def test_resample_in_time_linear_interpolation(self):
        input_cube = create_nx8x6_dataset_with_regular_coords(4)
        resampled_cube = resample_in_time(input_cube, "6h", interp_methods="linear")
        self.assertIn("time", resampled_cube)
        self.assertIn("refl", resampled_cube)
        self.assertIn("ndvi", resampled_cube)
        self.assertEqual((13,), resampled_cube.time.shape)
        np.testing.assert_allclose(
            resampled_cube.refl.values[..., 0, 1],
            np.array(
                [-1.0, -0.75, -0.5, -0.25, 0, 0.25, 0.5, 0.75, 1, 1.25, 1.5, 1.75, 2]
            ),
        )

    def test_resample_in_time_list_interpolation(self):
        input_cube = create_nx8x6_dataset_with_regular_coords(4)
        resampled_cube = resample_in_time(
            input_cube, "6h", interp_methods=["linear", "nearest"]
        )
        self.assertIn("time", resampled_cube)
        self.assertIn("refl_linear", resampled_cube)
        self.assertIn("refl_nearest", resampled_cube)
        self.assertIn("ndvi_linear", resampled_cube)
        self.assertIn("ndvi_nearest", resampled_cube)
        self.assertEqual((13,), resampled_cube.time.shape)
        np.testing.assert_allclose(
            resampled_cube.refl_linear.values[..., 0, 1],
            np.array(
                [-1.0, -0.75, -0.5, -0.25, 0, 0.25, 0.5, 0.75, 1, 1.25, 1.5, 1.75, 2]
            ),
        )

    def test_resample_in_time_variable_selection(self):
        input_cube = create_nx8x6_dataset_with_regular_coords(4)
        resampled_cube = resample_in_time(
            input_cube, "6h", interp_methods="linear", variables="refl"
        )
        self.assertIn("time", resampled_cube)
        self.assertIn("refl", resampled_cube)
        self.assertNotIn("ndvi", resampled_cube)
        self.assertEqual((13,), resampled_cube.time.shape)
        np.testing.assert_allclose(
            resampled_cube.refl.values[..., 0, 1],
            np.array(
                [-1.0, -0.75, -0.5, -0.25, 0, 0.25, 0.5, 0.75, 1, 1.25, 1.5, 1.75, 2]
            ),
        )

        resampled_cube = resample_in_time(
            input_cube, "6h", interp_methods="linear", variables="ndvi"
        )
        self.assertIn("time", resampled_cube)
        self.assertNotIn("refl", resampled_cube)
        self.assertIn("ndvi", resampled_cube)
        self.assertEqual((13,), resampled_cube.time.shape)
        np.testing.assert_allclose(
            resampled_cube.ndvi.values[..., 0, 1],
            np.array(
                [-1.0, -0.75, -0.5, -0.25, 0, 0.25, 0.5, 0.75, 1, 1.25, 1.5, 1.75, 2]
            ),
        )

    def test_resample_in_time_different_interp_method_per_variable(self):
        input_cube = create_nx8x6_dataset_with_regular_coords(4)
        resampled_cube = resample_in_time(
            input_cube,
            "6h",
            interp_methods={"refl": "linear", "ndvi": "nearest"},
        )
        self.assertIn("refl_linear", resampled_cube)
        self.assertIn("ndvi_nearest", resampled_cube)
        self.assertEqual((13,), resampled_cube.time.shape)
        np.testing.assert_allclose(
            resampled_cube.refl_linear.values[..., 0, 1],
            np.array(
                [-1.0, -0.75, -0.5, -0.25, 0, 0.25, 0.5, 0.75, 1, 1.25, 1.5, 1.75, 2]
            ),
        )
        np.testing.assert_allclose(
            resampled_cube.ndvi_nearest.values[..., 0, 1],
            np.array(
                [-1.0, -1.0, -1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0]
            ),
        )

    def test_resample_in_time_different_agg_method_per_variable(self):
        input_cube = create_nx8x6_dataset_with_regular_coords(4)
        resampled_cube = resample_in_time(
            input_cube, "2D", agg_methods={"refl": "max", "ndvi": "mean"}
        )
        self.assertIn("refl_max", resampled_cube)
        self.assertIn("ndvi_mean", resampled_cube)
        self.assertEqual((2,), resampled_cube.time.shape)
        np.testing.assert_allclose(
            resampled_cube.refl_max.values[..., 0, 1], np.array([0.0, 2.0])
        )
        np.testing.assert_allclose(
            resampled_cube.ndvi_mean.values[..., 0, 1], np.array([-0.5, 1.5])
        )

    def test_irregular_time_series_returns_orignal_dataset(self):
        with self.assertLogs("xcube.resampling", level="WARNING") as cm:
            resampled_cube = resample_in_time(self.ds_irregular, "2D")
        self.assertIn("Could not determine resampling operation.", cm.output[-1])
        self.assertEqual(resampled_cube, self.ds_irregular)
        self.assertEqual((4,), resampled_cube.time.shape)

    def test_irregular_time_series_agg(self):
        resampled_cube = resample_in_time(
            self.ds_irregular, "3D", agg_methods={"a": "max"}
        )
        self.assertIn("a_max", resampled_cube)
        self.assertNotIn("b_max", resampled_cube)
        self.assertEqual((3,), resampled_cube.time.shape)
        np.testing.assert_allclose(resampled_cube.a_max.values[..., 0], np.array([1]))

    def test_irregular_time_series_interp(self):
        resampled_cube = resample_in_time(
            self.ds_irregular, "12h", interp_methods={"a": "nearest"}
        )
        self.assertIn("a_nearest", resampled_cube)
        self.assertNotIn("b_nearest", resampled_cube)
        self.assertEqual((13,), resampled_cube.time.shape)
        np.testing.assert_allclose(
            resampled_cube.a_nearest.values[..., 0], np.array([0])
        )

    def test_no_time_dim(self):
        with self.assertRaises(ValueError) as cm:
            _ = resample_in_time(self.ds_notime, "1D")
        self.assertIn("Dataset must have a 'time' dimension.", str(cm.exception))

    def test_both_agg_and_interp_methods_raises(self):
        with self.assertRaises(ValueError):
            _guess_resampling_operation(
                self.ds_regular, "1D", interp_methods="linear", agg_methods="mean"
            )

    def test_too_few_time_points_raises(self):
        ds_single = xr.Dataset({"a": ("time", [1])}, coords={"time": ["2020-01-01"]})
        with self.assertRaises(ValueError):
            _guess_resampling_operation(ds_single, "1D")

    def test_explicit_agg_methods(self):
        result = _guess_resampling_operation(self.ds_regular, "1D", agg_methods="mean")
        self.assertEqual(result, "agg")

    def test_explicit_interp_methods(self):
        result = _guess_resampling_operation(
            self.ds_regular, "1D", interp_methods="linear"
        )
        self.assertEqual(result, "interp")

    def test_irregular_time_series_returns_none(self):
        result = _guess_resampling_operation(self.ds_irregular, "1D")
        self.assertIsNone(result)

    def test_regular_time_series_returns_none(self):
        result = _guess_resampling_operation(self.ds_regular, "1D")
        self.assertIsNone(result)

    def test_regular_time_series_downsample_returns_agg(self):
        result = _guess_resampling_operation(self.ds_regular, "2D")
        self.assertEqual(result, "agg")

    def test_regular_time_series_upsample_returns_interp(self):
        result = _guess_resampling_operation(self.ds_regular, "12h")
        self.assertEqual(result, "interp")

    def test_resample_in_time_invalid_method(self):
        input_cube = create_nx8x6_dataset_with_regular_coords(4)
        with self.assertRaises(ValueError):
            resample_in_time(input_cube, "6h", interp_methods=["nonlinear", "nearest"])

    def test_get_agg_method_kwargs(self):
        expected = {"tolerance": "1h"}
        result = _get_agg_method_kwargs("backfill", "1h")
        self.assertEqual(expected, result)

        expected = {"dim": "time", "keep_attrs": True}
        result = _get_agg_method_kwargs("all", "6h")
        self.assertEqual(expected, result)

        with self.assertRaises(ValueError):
            _get_agg_method_kwargs("bla", "6h")

    def test_get_temporal_interp_method(self):
        expected = ["nearest"]
        result = _get_temporal_interp_method(None, "a", xr.DataArray([1, 2]))
        self.assertEqual(expected, result)

        expected = ["linear"]
        result = _get_temporal_interp_method(None, "a", xr.DataArray([1.0, 2.0]))
        self.assertEqual(expected, result)

        expected = ["linear"]
        with self.assertLogs("xcube.resampling", level="WARNING") as cm:
            result = _get_temporal_interp_method(
                {"b": "linear"}, "a", xr.DataArray([1.0, 2.0])
            )
        self.assertIn("Interpolation method could not be derived ", cm.output[-1])
        self.assertEqual(expected, result)

    def test_get_temporal_agg_method(self):
        expected = ["nearest"]
        result = _get_temporal_agg_method(None, "a", xr.DataArray([1, 2]))
        self.assertEqual(expected, result)

        expected = ["mean"]
        result = _get_temporal_agg_method(None, "a", xr.DataArray([1.0, 2.0]))
        self.assertEqual(expected, result)

        expected = ["mean"]
        with self.assertLogs("xcube.resampling", level="WARNING") as cm:
            result = _get_temporal_agg_method(
                {"b": "linear"}, "a", xr.DataArray([1.0, 2.0])
            )
        self.assertIn("Aggregation method could not be derived ", cm.output[-1])
        self.assertEqual(expected, result)



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import pyproj
import xarray as xr

from .base import DEFAULT_TOLERANCE, GridMapping
from .cfconv import get_dataset_grid_mapping_proxies
from .coords import new_grid_mapping_from_coords
from .helpers import _normalize_crs


def new_grid_mapping_from_dataset(
    dataset: xr.Dataset,
    *,
    crs: str | pyproj.crs.CRS = None,
    tile_size: int | tuple[str, str] = None,
    prefer_crs: str | pyproj.crs.CRS = None,
    prefer_is_regular: bool = None,
    emit_warnings: bool = False,
    tolerance: float = DEFAULT_TOLERANCE,
) -> GridMapping:
    # Note `crs` is used if CRS is known in advance,
    # so the code forces its use. `prefer_crs` is used if
    # multiple CRSes are found, and a preference exists.
    # If it is not given, but `crs` is given,
    # then they are the same.
    if crs is not None:
        crs = _normalize_crs(crs)
    if prefer_crs is not None:
        prefer_crs = _normalize_crs(prefer_crs)
    else:
        prefer_crs = crs

    grid_mapping_proxies = get_dataset_grid_mapping_proxies(
        dataset,
        emit_warnings=emit_warnings,
        missing_projected_crs=crs,
        missing_rotated_latitude_longitude_crs=crs,
        missing_latitude_longitude_crs=crs,
    ).values()

    grid_mappings = [
        new_grid_mapping_from_coords(
            x_coords=gmp.coords.x,
            y_coords=gmp.coords.y,
            crs=gmp.crs,
            tile_size=tile_size or gmp.tile_size,
            tolerance=tolerance,
        )
        for gmp in grid_mapping_proxies
    ]

    if len(grid_mappings) > 1:
        if prefer_crs is not None and prefer_is_regular is not None:
            for gm in grid_mappings:
                if gm.crs == prefer_crs and bool(gm.is_regular) == prefer_is_regular:
                    return gm
            for gm in grid_mappings:
                if (
                    gm.crs.is_geographic
                    and prefer_crs.is_geographic
                    and bool(gm.is_regular) == prefer_is_regular
                ):
                    return gm

        if prefer_crs is not None:
            for gm in grid_mappings:
                if gm.crs == prefer_crs:
                    return gm
            for gm in grid_mappings:
                if gm.crs.is_geographic and prefer_crs.is_geographic:
                    return gm

        if prefer_is_regular is not None:
            for gm in grid_mappings:
                if bool(gm.is_regular) == prefer_is_regular:
                    return gm

    # Get arbitrary one (here: first)
    if grid_mappings:
        return grid_mappings[0]

    raise ValueError("cannot find any grid mapping in dataset")



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from collections.abc import Iterable

import xarray as xr

from .affine import affine_transform_dataset
from .constants import (
    LOG,
    FillValues,
    PreventNaNPropagations,
    SpatialAggMethods,
    SpatialInterpMethods,
)
from .gridmapping import GridMapping
from .rectify import rectify_dataset
from .reproject import reproject_dataset
from .utils import _can_apply_affine_transform


def resample_in_space(
    source_ds: xr.Dataset,
    *,
    target_gm: GridMapping | None = None,
    source_gm: GridMapping | None = None,
    variables: str | Iterable[str] | None = None,
    interp_methods: SpatialInterpMethods | None = None,
    agg_methods: SpatialAggMethods | None = None,
    prevent_nan_propagations: PreventNaNPropagations = False,
    fill_values: FillValues | None = None,
    tile_size: int | tuple[int, int] | None = None,
) -> xr.Dataset:
    """
    Resample the spatial dimensions of a dataset to match a target grid mapping.

    Depending on the regularity and compatibility of the source and target grid
    mappings, this function will either rectify, reproject, or affine-transform the
    spatial dimensions of `source_ds`.

    Args:
        source_ds: The input xarray.Dataset. Data variables must have dimensions
            in the following order: optional third dimension followed by the
            y-dimension (e.g., "y" or "lat") and the x-dimension (e.g., "x" or "lon").
        target_gm: The target GridMapping to which the dataset should be resampled.
            Must be regular. If not provided, a default regular grid is derived
            from `source_gm` using `to_regular(tile_size)`.
        source_gm: The GridMapping describing the source dataset's spatial layout.
            If not provided, it is inferred from `source_ds` using
            `GridMapping.from_dataset(source_ds)`.
        variables: A single variable name or iterable of variable names to be
            resampled. If None, all data variables will be processed.
        interp_methods: Optional interpolation method to be used for upsampling spatial
            data variables. Can be a single interpolation method for all variables or a
            dictionary mapping variable names or dtypes to interpolation method.
            Supported methods include:

            - `0` (nearest neighbor)
            - `1` (linear / bilinear)
            - `"nearest"`
            - `"triangular"`
            - `"bilinear"`

            The default is `0` for integer arrays, else `1`.
        agg_methods: Optional aggregation methods for downsampling spatial variables.
            Can be a single method for all variables or a dictionary mapping variable
            names or dtypes to methods. Supported methods include:
                "center", "count", "first", "last", "max", "mean", "median",
                "mode", "min", "prod", "std", "sum", and "var".
            Defaults to "center" for integer arrays, else "mean".
        prevent_nan_propagations: Optional boolean or mapping to prevent NaN
            propagation during upsampling (only applies when interpolation method
            is not nearest). Can be a single boolean or a dictionary mapping
            variable names or dtypes to booleans. Defaults to False.
        fill_values: Optional fill value(s) for areas outside input coverage.
            Can be a single value or dictionary by variable or type. If not provided,
            defaults based on data type are used:

            - float: NaN
            - uint8: 255
            - uint16: 65535
            - other ints: -1

        tile_size: Optional tile size used when generating a regular grid from
            an irregular source grid mapping. Only used if `target_gm` is not provided.

    Returns:
        A new dataset that has been spatially resampled to match the target grid
            mapping.

    Notes:
        - If `source_gm` is not provided, it is inferred from `source_ds`.
        - If `target_gm` is not provided, and the source is irregular, it is
          derived from `source_gm.to_regular(tile_size=tile_size)`.
        - If both grid mappings are regular and approximately equal, the original
          dataset is returned unchanged.
        - If the transformation can be represented as an affine mapping, it is
          applied directly for performance.
        - If the source is irregular, rectification is applied.
        - Otherwise, a reprojection is performed.
        - See the [xcube-resampling documentation](https://xcube-dev.github.io/xcube-resampling/)
          for more details.
    """
    if source_gm is None:
        source_gm = GridMapping.from_dataset(source_ds)

    if not source_gm.is_regular:
        return rectify_dataset(
            source_ds,
            target_gm=target_gm,
            source_gm=source_gm,
            variables=variables,
            interp_methods=interp_methods,
            agg_methods=agg_methods,
            prevent_nan_propagations=prevent_nan_propagations,
            fill_values=fill_values,
            tile_size=tile_size,
        )
    else:
        if target_gm is None:
            LOG.warning(
                "If source grid mapping is regular `target_gm` must be given. "
                "Source dataset is returned."
            )
            return source_ds
        GridMapping.assert_regular(target_gm, name="target_gm")
        if source_gm.is_close(target_gm):
            return source_ds

        if _can_apply_affine_transform(source_gm, target_gm):
            return affine_transform_dataset(
                source_ds,
                target_gm,
                source_gm=source_gm,
                variables=variables,
                interp_methods=interp_methods,
                agg_methods=agg_methods,
                prevent_nan_propagations=prevent_nan_propagations,
                fill_values=fill_values,
            )
        else:
            return reproject_dataset(
                source_ds,
                target_gm,
                source_gm=source_gm,
                variables=variables,
                interp_methods=interp_methods,
                agg_methods=agg_methods,
                prevent_nan_propagations=prevent_nan_propagations,
                fill_values=fill_values,
            )



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import shutil
import unittest
import warnings

import numpy as np
import pyproj
import xarray as xr

# noinspection PyProtectedMember
from xcube_resampling.gridmapping.cfconv import (
    GridCoords,
    GridMappingProxy,
    _find_potential_coord_vars,
    _is_potential_coord_var,
    get_dataset_grid_mapping_proxies,
)

CRS_WGS84 = pyproj.crs.CRS(4326)
CRS_CRS84 = pyproj.crs.CRS.from_string("urn:ogc:def:crs:OGC:1.3:CRS84")
CRS_UTM_33N = pyproj.crs.CRS(32633)

CRS_ROTATED_POLE = pyproj.crs.CRS.from_cf(
    dict(
        grid_mapping_name="rotated_latitude_longitude",
        grid_north_pole_latitude=32.5,
        grid_north_pole_longitude=170.0,
    )
)


class GetDatasetGridMappingsTest(unittest.TestCase):
    def test_no_crs_lon_lat_common_names(self):
        dataset = xr.Dataset(
            coords=dict(
                lon=xr.DataArray(np.linspace(10, 12, 11), dims="lon"),
                lat=xr.DataArray(np.linspace(50, 52, 11), dims="lat"),
            )
        )
        grid_mappings = get_dataset_grid_mapping_proxies(dataset)
        self.assertEqual(1, len(grid_mappings))
        self.assertIn(None, grid_mappings)
        grid_mapping = grid_mappings.get(None)
        self.assertIsInstance(grid_mapping, GridMappingProxy)
        self.assertEqual(CRS_WGS84, grid_mapping.crs)
        self.assertEqual("latitude_longitude", grid_mapping.name)
        self.assertIsInstance(grid_mapping.coords, GridCoords)
        self.assertIsInstance(grid_mapping.coords.x, xr.DataArray)
        self.assertIsInstance(grid_mapping.coords.y, xr.DataArray)
        self.assertEqual("lon", grid_mapping.coords.x.name)
        self.assertEqual("lat", grid_mapping.coords.y.name)

    def test_no_crs_lon_lat_standard_names(self):
        dataset = xr.Dataset(
            coords=dict(
                weird_x=xr.DataArray(
                    np.linspace(10, 12, 11),
                    dims="i",
                    attrs=dict(standard_name="longitude"),
                ),
                weird_y=xr.DataArray(
                    np.linspace(50, 52, 11),
                    dims="j",
                    attrs=dict(standard_name="latitude"),
                ),
            )
        )
        grid_mappings = get_dataset_grid_mapping_proxies(dataset)
        self.assertEqual(1, len(grid_mappings))
        self.assertIn(None, grid_mappings)
        grid_mapping = grid_mappings.get(None)
        self.assertIsInstance(grid_mapping, GridMappingProxy)
        self.assertEqual(CRS_WGS84, grid_mapping.crs)
        self.assertEqual("latitude_longitude", grid_mapping.name)
        self.assertIsInstance(grid_mapping.coords, GridCoords)
        self.assertIsInstance(grid_mapping.coords.x, xr.DataArray)
        self.assertIsInstance(grid_mapping.coords.y, xr.DataArray)
        self.assertEqual("weird_x", grid_mapping.coords.x.name)
        self.assertEqual("weird_y", grid_mapping.coords.y.name)

    def test_crs_x_y_with_common_names(self):
        dataset = xr.Dataset(
            dict(crs=xr.DataArray(0, attrs=CRS_UTM_33N.to_cf())),
            coords=dict(
                x=xr.DataArray(np.linspace(1000, 12000, 11), dims="x"),
                y=xr.DataArray(np.linspace(5000, 52000, 11), dims="y"),
            ),
        )
        grid_mappings = get_dataset_grid_mapping_proxies(dataset)
        self.assertEqual(1, len(grid_mappings))
        self.assertIn("crs", grid_mappings)
        grid_mapping = grid_mappings.get("crs")
        self.assertIsInstance(grid_mapping, GridMappingProxy)
        self.assertEqual(CRS_UTM_33N, grid_mapping.crs)
        self.assertEqual("transverse_mercator", grid_mapping.name)
        self.assertIsInstance(grid_mapping.coords, GridCoords)
        self.assertIsInstance(grid_mapping.coords.x, xr.DataArray)
        self.assertIsInstance(grid_mapping.coords.y, xr.DataArray)
        self.assertEqual("x", grid_mapping.coords.x.name)
        self.assertEqual("y", grid_mapping.coords.y.name)

    def test_crs_x_y_with_standard_names(self):
        dataset = xr.Dataset(
            dict(crs=xr.DataArray(0, attrs=CRS_UTM_33N.to_cf())),
            coords=dict(
                myx=xr.DataArray(
                    np.linspace(1000, 12000, 11),
                    dims="x",
                    attrs=dict(standard_name="projection_x_coordinate"),
                ),
                myy=xr.DataArray(
                    np.linspace(5000, 52000, 11),
                    dims="y",
                    attrs=dict(standard_name="projection_y_coordinate"),
                ),
            ),
        )
        grid_mappings = get_dataset_grid_mapping_proxies(dataset)
        self.assertEqual(1, len(grid_mappings))
        self.assertIn("crs", grid_mappings)
        grid_mapping = grid_mappings.get("crs")
        self.assertIsInstance(grid_mapping, GridMappingProxy)
        self.assertEqual(CRS_UTM_33N, grid_mapping.crs)
        self.assertEqual("transverse_mercator", grid_mapping.name)
        self.assertIsInstance(grid_mapping.coords, GridCoords)
        self.assertIsInstance(grid_mapping.coords.x, xr.DataArray)
        self.assertIsInstance(grid_mapping.coords.y, xr.DataArray)
        self.assertEqual("myx", grid_mapping.coords.x.name)
        self.assertEqual("myy", grid_mapping.coords.y.name)

    def test_latitude_longitude_with_x_y(self):
        # This is what we get when opening a CRS-84 GeoTIFF using rioxarray
        dataset = xr.Dataset(
            dict(
                band_1=xr.DataArray(np.zeros((11, 11)), dims=["y", "x"]),
                spatial_ref=xr.DataArray(0, attrs=CRS_CRS84.to_cf()),
            ),
            coords=dict(
                x=xr.DataArray(np.linspace(10, 20, 11), dims="x"),
                y=xr.DataArray(np.linspace(50, 40, 11), dims="y"),
            ),
        )
        grid_mappings = get_dataset_grid_mapping_proxies(dataset)
        self.assertEqual(1, len(grid_mappings))
        self.assertIn("spatial_ref", grid_mappings)
        grid_mapping = grid_mappings.get("spatial_ref")
        self.assertIsInstance(grid_mapping, GridMappingProxy)
        self.assertEqual(CRS_CRS84, grid_mapping.crs)
        self.assertEqual("latitude_longitude", grid_mapping.name)
        self.assertIsInstance(grid_mapping.coords, GridCoords)
        self.assertIsInstance(grid_mapping.coords.x, xr.DataArray)
        self.assertIsInstance(grid_mapping.coords.y, xr.DataArray)
        self.assertEqual("x", grid_mapping.coords.x.name)
        self.assertEqual("y", grid_mapping.coords.y.name)

    def test_crs_in_attrs(self):
        dataset = xr.Dataset(
            coords=dict(
                lon=xr.DataArray(np.linspace(10, 12, 11), dims="lon"),
                lat=xr.DataArray(np.linspace(50, 52, 11), dims="lat"),
            ),
            attrs={
                "crs_wkt": (
                    'GEOGCRS["WGS 84",ENSEMBLE["World Geodetic System 1984 ensemble",'
                    'MEMBER["World Geodetic System 1984 (Transit)"],MEMBER["World '
                    'Geodetic System 1984 (G730)"],MEMBER["World Geodetic System 1984 '
                    '(G873)"],MEMBER["World Geodetic System 1984 (G1150)"],MEMBER'
                    '["World Geodetic System 1984 (G1674)"],MEMBER["World Geodetic '
                    'System 1984 (G1762)"],MEMBER["World Geodetic System 1984 (G2139)"]'
                    ',MEMBER["World Geodetic System 1984 (G2296)"],ELLIPSOID'
                    '["WGS 84",6378137,298.257223563,LENGTHUNIT["metre",1]],'
                    'ENSEMBLEACCURACY[2.0]],PRIMEM["Greenwich",0,ANGLEUNIT'
                    '["degree",0.0174532925199433]],CS[ellipsoidal,2],AXIS'
                    '["geodetic latitude (Lat)",north,ORDER[1],ANGLEUNIT'
                    '["degree",0.0174532925199433]],AXIS["geodetic longitude (Lon)"'
                    ',east,ORDER[2],ANGLEUNIT["degree",0.0174532925199433]],'
                    'USAGE[SCOPE["Horizontal component of 3D system."],AREA["World."]'
                    ',BBOX[-90,-180,90,180]],ID["EPSG",4326]]'
                ),
                "semi_major_axis": 6378137.0,
                "semi_minor_axis": 6356752.314245179,
                "inverse_flattening": 298.257223563,
                "reference_ellipsoid_name": "WGS 84",
                "longitude_of_prime_meridian": 0.0,
                "prime_meridian_name": "Greenwich",
                "geographic_crs_name": "WGS 84",
                "horizontal_datum_name": "World Geodetic System 1984 ensemble",
                "grid_mapping_name": "latitude_longitude",
            },
        )
        grid_mappings = get_dataset_grid_mapping_proxies(dataset)
        self.assertEqual(1, len(grid_mappings))
        self.assertIn(None, grid_mappings)
        grid_mapping = grid_mappings.get(None)
        self.assertIsInstance(grid_mapping, GridMappingProxy)
        self.assertEqual(CRS_WGS84, grid_mapping.crs)
        self.assertEqual("latitude_longitude", grid_mapping.name)
        self.assertIsInstance(grid_mapping.coords, GridCoords)
        self.assertIsInstance(grid_mapping.coords.x, xr.DataArray)
        self.assertIsInstance(grid_mapping.coords.y, xr.DataArray)
        self.assertEqual("lon", grid_mapping.coords.x.name)
        self.assertEqual("lat", grid_mapping.coords.y.name)

    def test_emit_warning(self):
        dataset = xr.Dataset(
            coords=dict(
                lon=xr.DataArray([10], dims="lon"),
                lat=xr.DataArray([50], dims="lat"),
            ),
        )
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            _ = get_dataset_grid_mapping_proxies(dataset, emit_warnings=True)
        self.assertEqual(len(w), 1)
        self.assertIn("missing x- and/or y-coordinates", str(w[0].message))

    def test_rotated_pole_with_common_names(self):
        dataset = xr.Dataset(
            dict(rotated_pole=xr.DataArray(0, attrs=CRS_ROTATED_POLE.to_cf())),
            coords=dict(
                rlon=xr.DataArray(np.linspace(-180, 180, 11), dims="rlon"),
                rlat=xr.DataArray(np.linspace(0, 90, 11), dims="rlat"),
            ),
        )
        grid_mappings = get_dataset_grid_mapping_proxies(dataset)
        self.assertEqual(1, len(grid_mappings))
        self.assertIn("rotated_pole", grid_mappings)
        grid_mapping = grid_mappings.get("rotated_pole")
        self.assertIsInstance(grid_mapping, GridMappingProxy)
        self.assertIn("Geographic", grid_mapping.crs.type_name)
        self.assertIsInstance(grid_mapping.coords, GridCoords)
        self.assertIsInstance(grid_mapping.coords.x, xr.DataArray)
        self.assertIsInstance(grid_mapping.coords.y, xr.DataArray)
        self.assertEqual("rlon", grid_mapping.coords.x.name)
        self.assertEqual("rlat", grid_mapping.coords.y.name)

    def test_rotated_pole_with_standard_names(self):
        dataset = xr.Dataset(
            dict(rotated_pole=xr.DataArray(0, attrs=CRS_ROTATED_POLE.to_cf())),
            coords=dict(
                u=xr.DataArray(
                    np.linspace(-180, 180, 11),
                    dims="u",
                    attrs=dict(standard_name="grid_longitude"),
                ),
                v=xr.DataArray(
                    np.linspace(0, 90, 11),
                    dims="v",
                    attrs=dict(standard_name="grid_latitude"),
                ),
            ),
        )
        grid_mappings = get_dataset_grid_mapping_proxies(dataset)
        self.assertEqual(1, len(grid_mappings))
        self.assertIn("rotated_pole", grid_mappings)
        grid_mapping = grid_mappings.get("rotated_pole")
        self.assertIsInstance(grid_mapping, GridMappingProxy)
        self.assertIn("Geographic", grid_mapping.crs.type_name)
        self.assertIsInstance(grid_mapping.coords, GridCoords)
        self.assertIsInstance(grid_mapping.coords.x, xr.DataArray)
        self.assertIsInstance(grid_mapping.coords.y, xr.DataArray)
        self.assertEqual("u", grid_mapping.coords.x.name)
        self.assertEqual("v", grid_mapping.coords.y.name)

    def test_bounds_detection(self):
        dataset = xr.Dataset(
            coords={
                "lon": xr.DataArray(np.linspace(0, 10, 5), dims="lon"),
                "lat": xr.DataArray(np.linspace(0, 5, 5), dims="lat"),
                "lon_bnds": xr.DataArray(np.linspace(0, 10, 10), dims="bnds"),
                "lat_bounds": xr.DataArray(np.linspace(0, 5, 10), dims="bnds"),
                "alt": xr.DataArray(np.linspace(0, 100, 5), dims="alt"),
            }
        )
        dataset["lat"].attrs["bounds"] = "lat_bounds"

        potential_vars = _find_potential_coord_vars(dataset)

        # lon_bnds and lat_bounds should be excluded, but lon/lat/alt included
        self.assertIn("lon", potential_vars)
        self.assertIn("lat", potential_vars)
        self.assertIn("alt", potential_vars)
        self.assertNotIn("lon_bnds", potential_vars)
        self.assertNotIn("lat_bounds", potential_vars)

    def test_coordinates_in_attribute(self):
        dataset = xr.Dataset(
            {
                "x": xr.DataArray([0, 1]),
                "y": xr.DataArray([0, 1]),
            },
            attrs={"coordinates": "x y"},
        )

        result = _find_potential_coord_vars(dataset)
        self.assertIn("x", result)
        self.assertIn("y", result)

    def test_var_not_in_dataset(self):
        dataset = xr.Dataset(
            coords={
                "lon": xr.DataArray(np.linspace(0, 10, 5), dims="lon"),
                "lat": xr.DataArray(np.linspace(0, 5, 5), dims="lat"),
            }
        )
        bounds_vars = set()
        self.assertFalse(_is_potential_coord_var(dataset, bounds_vars, "missing_var"))


class XarrayDecodeCfTest(unittest.TestCase):
    """Find out how xarray treats 1D and 2D coordinate variables when decode_cf=True or =False"""

    def test_cf_1d_coords(self):
        self._write_coords(*self._gen_1d())
        self.assertVarNames({"noise", "crs"}, {"lon", "lat"}, decode_cf=True)
        self.assertVarNames({"noise", "crs"}, {"lon", "lat"}, decode_cf=False)

    def test_cf_1d_data_vars(self):
        self._write_data_vars(*self._gen_1d())
        self.assertVarNames({"noise", "crs"}, {"lon", "lat"}, decode_cf=True)
        self.assertVarNames({"noise", "crs"}, {"lon", "lat"}, decode_cf=False)

    def test_cf_2d_coords(self):
        self._write_coords(*self._gen_2d())
        self.assertVarNames({"noise", "crs"}, {"lon", "lat"}, decode_cf=True)
        self.assertVarNames({"noise", "crs", "lon", "lat"}, set(), decode_cf=False)

    def test_cf_2d_data_vars(self):
        self._write_data_vars(*self._gen_2d())
        self.assertVarNames({"noise", "crs", "lon", "lat"}, set(), decode_cf=True)
        self.assertVarNames({"noise", "crs", "lon", "lat"}, set(), decode_cf=False)

    def assertVarNames(
        self, exp_data_vars: set[str], exp_coords: set[str], decode_cf: bool
    ):
        ds1 = xr.open_zarr("noise.zarr", decode_cf=decode_cf)
        self.assertEqual(exp_coords, set(ds1.coords))
        self.assertEqual(exp_data_vars, set(ds1.data_vars))

    @classmethod
    def _write_coords(cls, noise, crs, lon, lat):
        dataset = xr.Dataset(dict(noise=noise, crs=crs), coords=dict(lon=lon, lat=lat))
        dataset.to_zarr("noise.zarr", mode="w")

    @classmethod
    def _write_data_vars(cls, noise, crs, lon, lat):
        dataset = xr.Dataset(dict(noise=noise, crs=crs, lon=lon, lat=lat))
        dataset.to_zarr("noise.zarr", mode="w")

    @classmethod
    def tearDownClass(cls) -> None:
        shutil.rmtree("noise.zarr")

    def _gen_1d(self):
        noise = xr.DataArray(np.random.random((11, 11)), dims=("lat", "lon"))
        crs = xr.DataArray(0, attrs=CRS_CRS84.to_cf())
        lon = xr.DataArray(np.linspace(10, 12, 11), dims="lon")
        lat = xr.DataArray(np.linspace(50, 52, 11), dims="lat")
        noise.attrs["grid_mapping"] = "crs"
        lon.attrs["standard_name"] = "longitude"
        lat.attrs["standard_name"] = "latitude"
        self.assertEqual(("lon",), lon.dims)
        self.assertEqual(("lat",), lat.dims)
        return noise, crs, lon, lat

    def _gen_2d(self):
        noise = xr.DataArray(np.random.random((11, 11)), dims=("y", "x"))
        crs = xr.DataArray(0, attrs=CRS_CRS84.to_cf())
        lon = xr.DataArray(np.linspace(10, 12, 11), dims="x")
        lat = xr.DataArray(np.linspace(50, 52, 11), dims="y")
        lat, lon = xr.broadcast(lat, lon)
        noise.attrs["grid_mapping"] = "crs"
        lon.attrs["standard_name"] = "longitude"
        lat.attrs["standard_name"] = "latitude"
        self.assertEqual(("y", "x"), lon.dims)
        self.assertEqual(("y", "x"), lat.dims)
        return noise, crs, lon, lat



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import unittest

import dask.array as da
import numpy as np
import pyproj
import xarray as xr

from xcube_resampling.gridmapping import GridMapping

# noinspection PyProtectedMember

GEO_CRS = pyproj.crs.CRS(4326)
NOT_A_GEO_CRS = pyproj.crs.CRS(5243)


# noinspection PyMethodMayBeStatic
class Coords1DGridMappingTest(unittest.TestCase):
    def test_1d_j_axis_down(self):
        x_coords = xr.DataArray(np.linspace(1.5, 8.5, 8), dims="lon")
        y_coords = xr.DataArray(np.linspace(4.5, -4.5, 10), dims="lat")
        gm = GridMapping.from_coords(x_coords=x_coords, y_coords=y_coords, crs=GEO_CRS)
        self.assertEqual((8, 10), gm.size)
        self.assertEqual((8, 10), gm.tile_size)
        self.assertEqual((1, 1), gm.xy_res)
        self.assertEqual((1, -5, 9, 5), gm.xy_bbox)
        self.assertEqual(GEO_CRS, gm.crs)
        self.assertEqual(True, gm.is_regular)
        self.assertEqual(False, gm.is_j_axis_up)
        self.assertEqual(False, gm.is_lon_360)
        self.assertTrue(hasattr(gm, "x_coords"))
        self.assertTrue(hasattr(gm, "y_coords"))
        # noinspection PyUnresolvedReferences
        self.assertIs(x_coords, gm.x_coords)
        # noinspection PyUnresolvedReferences
        self.assertIs(y_coords, gm.y_coords)
        self.assertIs(x_coords, gm._new_x_coords())
        self.assertIs(y_coords, gm._new_y_coords())

    def test_1d_j_axis_up(self):
        gm = GridMapping.from_coords(
            x_coords=xr.DataArray(np.linspace(1.5, 8.5, 8), dims="lon"),
            y_coords=xr.DataArray(np.linspace(-4.5, 4.5, 10), dims="lat"),
            crs=GEO_CRS,
        )
        self.assertEqual((8, 10), gm.size)
        self.assertEqual((8, 10), gm.tile_size)
        self.assertEqual((1, 1), gm.xy_res)
        self.assertEqual((1, -5, 9, 5), gm.xy_bbox)
        self.assertEqual(GEO_CRS, gm.crs)
        self.assertEqual(True, gm.is_regular)
        self.assertEqual(True, gm.is_j_axis_up)
        self.assertEqual(False, gm.is_lon_360)

    def test_1d_lon_360(self):
        gm = GridMapping.from_coords(
            x_coords=xr.DataArray(np.linspace(177.5, 184.5, 8), dims="lon"),
            y_coords=xr.DataArray(np.linspace(4.5, -4.5, 10), dims="lat"),
            crs=GEO_CRS,
        )
        self.assertEqual((8, 10), gm.size)
        self.assertEqual((8, 10), gm.tile_size)
        self.assertEqual((1, 1), gm.xy_res)
        self.assertEqual((177, -5, 185, 5), gm.xy_bbox)
        self.assertEqual(GEO_CRS, gm.crs)
        self.assertEqual(True, gm.is_regular)
        self.assertEqual(False, gm.is_j_axis_up)
        self.assertEqual(True, gm.is_lon_360)

    def test_1d_anti_meridian(self):
        lon = np.linspace(177.5, 184.5, 8)
        lon = np.where(lon > 180, lon - 360, lon)
        gm = GridMapping.from_coords(
            x_coords=xr.DataArray(lon, dims="lon"),
            y_coords=xr.DataArray(np.linspace(4.5, -4.5, 10), dims="lat"),
            crs=GEO_CRS,
        )
        self.assertEqual((8, 10), gm.size)
        self.assertEqual((8, 10), gm.tile_size)
        self.assertEqual((1, 1), gm.xy_res)
        self.assertEqual((177, -5, 185, 5), gm.xy_bbox)
        self.assertEqual(GEO_CRS, gm.crs)
        self.assertEqual(True, gm.is_regular)
        self.assertEqual(False, gm.is_j_axis_up)
        self.assertEqual(True, gm.is_lon_360)

    def test_1d_tiles_given(self):
        gm = GridMapping.from_coords(
            x_coords=xr.DataArray(np.linspace(177.5, 184.5, 8), dims="lon"),
            y_coords=xr.DataArray(np.linspace(4.5, -4.5, 10), dims="lat"),
            crs=GEO_CRS,
            tile_size=(5, 3),
        )
        self.assertEqual((8, 10), gm.size)
        self.assertEqual((5, 3), gm.tile_size)
        self.assertEqual((1, 1), gm.xy_res)
        self.assertEqual((177, -5, 185, 5), gm.xy_bbox)
        self.assertEqual(GEO_CRS, gm.crs)
        self.assertEqual(True, gm.is_regular)
        self.assertEqual(False, gm.is_j_axis_up)
        self.assertEqual(True, gm.is_lon_360)

    def test_1d_tiles_from_coords_chunks(self):
        gm = GridMapping.from_coords(
            x_coords=xr.DataArray(da.linspace(177.5, 184.5, 8, chunks=4), dims="lon"),
            y_coords=xr.DataArray(da.linspace(4.5, -4.5, 10, chunks=5), dims="lat"),
            crs=GEO_CRS,
        )
        self.assertEqual((8, 10), gm.size)
        self.assertEqual((4, 5), gm.tile_size)
        self.assertEqual((1, 1), gm.xy_res)
        self.assertEqual((177, -5, 185, 5), gm.xy_bbox)
        self.assertEqual(GEO_CRS, gm.crs)
        self.assertEqual(True, gm.is_regular)
        self.assertEqual(False, gm.is_j_axis_up)
        self.assertEqual(True, gm.is_lon_360)

    def test_1d_x_irregular(self):
        gm = GridMapping.from_coords(
            x_coords=xr.DataArray(
                [1.5, 2.5, 3.5, 4.5, 5.49, 6.5, 7.5, 8.5], dims="lon"
            ),
            y_coords=xr.DataArray(np.linspace(4.5, -4.5, 10), dims="lat"),
            crs=GEO_CRS,
        )
        self.assertEqual((8, 10), gm.size)
        self.assertEqual((8, 10), gm.tile_size)
        self.assertEqual((1, 1), gm.xy_res)
        self.assertEqual((1, -5, 9, 5), gm.xy_bbox)
        self.assertEqual(GEO_CRS, gm.crs)
        self.assertEqual(False, gm.is_regular)
        self.assertEqual(False, gm.is_j_axis_up)
        self.assertEqual(False, gm.is_lon_360)

    def test_1d_xy_coords(self):
        gm = GridMapping.from_coords(
            x_coords=xr.DataArray(np.linspace(1.5, 8.5, 8), dims="lon"),
            y_coords=xr.DataArray(np.linspace(4.5, -4.5, 10), dims="lat"),
            crs=GEO_CRS,
        )
        xy_coords = gm.xy_coords
        self.assertIsInstance(xy_coords, xr.DataArray)
        self.assertIs(xy_coords, gm.xy_coords)
        self.assertEqual(("coord", "lat", "lon"), xy_coords.dims)
        self.assertEqual((2, 10, 8), xy_coords.shape)
        self.assertEqual(("lon", "lat"), gm.xy_var_names)
        self.assertEqual(("lon", "lat"), gm.xy_dim_names)

    def test_to_coords(self):
        gm = GridMapping.regular(
            size=(10, 6), xy_min=(-2600.0, 1200.0), xy_res=10.0, crs=NOT_A_GEO_CRS
        )
        cv = gm.to_coords(reuse_coords=False)
        self.assertIn("x", cv)
        self.assertIn("y", cv)
        self.assertEqual(np.float64, cv["x"].dtype)
        self.assertEqual(np.float64, cv["y"].dtype)

        gm2 = GridMapping.from_coords(
            cv["x"].astype(np.float32), cv["y"].astype(np.float32), gm.crs
        )
        cv2 = gm2.to_coords(
            xy_var_names=("a", "b"), xy_dim_names=("u", "v"), reuse_coords=True
        )
        self.assertIn("a", cv2)
        self.assertIn("b", cv2)
        self.assertEqual(np.dtype(np.float32), cv2["a"].dtype)
        self.assertEqual(np.dtype(np.float32), cv2["b"].dtype)


class Coords2DGridMappingTest(unittest.TestCase):
    def test_2d(self):
        x_coords = xr.DataArray(
            [
                [10.0, 10.1, 10.2, 10.3],
                [10.1, 10.2, 10.3, 10.4],
                [10.2, 10.3, 10.4, 10.5],
            ],
            dims=("lat", "lon"),
        )
        y_coords = xr.DataArray(
            [
                [52.0, 52.2, 52.4, 52.6],
                [52.2, 52.4, 52.6, 52.8],
                [52.4, 52.6, 52.8, 53.0],
            ],
            dims=("lat", "lon"),
        )
        gm = GridMapping.from_coords(x_coords=x_coords, y_coords=y_coords, crs=GEO_CRS)
        self.assertEqual((4, 3), gm.size)
        self.assertEqual((4, 3), gm.tile_size)
        self.assertEqual((0.1, 0.2), gm.xy_res)
        self.assertEqual((9.95, 51.9, 10.55, 53.1), gm.xy_bbox)
        self.assertEqual(GEO_CRS, gm.crs)
        self.assertEqual(False, gm.is_regular)
        self.assertEqual(True, gm.is_j_axis_up)
        self.assertEqual(False, gm.is_lon_360)
        self.assertTrue(hasattr(gm, "x_coords"))
        self.assertTrue(hasattr(gm, "y_coords"))
        # noinspection PyUnresolvedReferences
        self.assertIs(x_coords, gm.x_coords)
        # noinspection PyUnresolvedReferences
        self.assertIs(y_coords, gm.y_coords)

    def test_2d_tile_size_from_chunks(self):
        gm = GridMapping.from_coords(
            x_coords=xr.DataArray(
                da.array(
                    [
                        [10.0, 10.1, 10.2, 10.3],
                        [10.1, 10.2, 10.3, 10.4],
                        [10.2, 10.3, 10.4, 10.5],
                    ]
                ).rechunk((2, 3)),
                dims=("lat", "lon"),
            ),
            y_coords=xr.DataArray(
                da.array(
                    [
                        [52.0, 52.2, 52.4, 52.6],
                        [52.2, 52.4, 52.6, 52.8],
                        [52.4, 52.6, 52.8, 53.0],
                    ]
                ).rechunk((2, 3)),
                dims=("lat", "lon"),
            ),
            crs=GEO_CRS,
        )
        self.assertEqual((4, 3), gm.size)
        self.assertEqual((3, 2), gm.tile_size)

    def test_2d_regular(self):
        gm = GridMapping.from_coords(
            x_coords=xr.DataArray(
                [
                    [10.2, 10.3, 10.4, 10.5],
                    [10.2, 10.3, 10.4, 10.5],
                    [10.2, 10.3, 10.4, 10.5],
                ],
                dims=("lat", "lon"),
            ),
            y_coords=xr.DataArray(
                [
                    [52.4, 52.4, 52.4, 52.4],
                    [52.6, 52.6, 52.6, 52.6],
                    [52.8, 52.8, 52.8, 52.8],
                ],
                dims=("lat", "lon"),
            ),
            crs=GEO_CRS,
        )
        self.assertEqual((4, 3), gm.size)
        self.assertEqual((4, 3), gm.tile_size)
        self.assertAlmostEqual(0.1, gm.x_res)
        self.assertAlmostEqual(0.2, gm.y_res)
        self.assertAlmostEqual(10.2, gm.x_min)
        self.assertAlmostEqual(52.4, gm.y_min)
        self.assertAlmostEqual(10.5, gm.x_max)
        self.assertAlmostEqual(52.8, gm.y_max)
        self.assertEqual(GEO_CRS, gm.crs)
        self.assertEqual(True, gm.is_regular)
        self.assertEqual(True, gm.is_j_axis_up)
        self.assertEqual(False, gm.is_lon_360)

    def test_2d_anti_meridian(self):
        gm = GridMapping.from_coords(
            x_coords=xr.DataArray(
                [
                    [+177.5, +178.5, +179.5, -179.5],
                    [+178.5, +179.5, -179.5, -178.5],
                    [+179.5, -179.5, -178.5, -177.5],
                ],
                dims=("lat", "lon"),
            ),
            y_coords=xr.DataArray(
                [
                    [52.4, 52.4, 52.4, 52.4],
                    [52.6, 52.6, 52.6, 52.6],
                    [52.8, 52.8, 52.8, 52.8],
                ],
                dims=("lat", "lon"),
            ),
            crs=GEO_CRS,
        )
        self.assertEqual((4, 3), gm.size)
        self.assertEqual((4, 3), gm.tile_size)
        self.assertAlmostEqual(1, gm.x_res)
        self.assertAlmostEqual(0.2, gm.y_res)
        self.assertEqual((177.0, 52.3, 183.0, 52.9), gm.xy_bbox)
        self.assertEqual(GEO_CRS, gm.crs)
        self.assertEqual(False, gm.is_regular)
        self.assertEqual(True, gm.is_j_axis_up)
        self.assertEqual(True, gm.is_lon_360)

    def test_to_regular(self):
        lon = xr.DataArray([[1.0, 6.0], [0.0, 2.0]], dims=("y", "x"))
        lat = xr.DataArray([[56.0, 53.0], [52.0, 50.0]], dims=("y", "x"))

        gm_irr = GridMapping.from_coords(lon, lat, GEO_CRS)
        gm_reg_actual = gm_irr.to_regular()
        gm_reg_expected = GridMapping.regular(
            size=(3, 3),
            tile_size=(2, 2),
            xy_min=(0, 50),
            xy_res=3.5,
            crs=GEO_CRS,
        )
        self.assertEqual(gm_reg_expected.size, gm_reg_actual.size)
        self.assertEqual(gm_reg_expected.tile_size, gm_reg_actual.tile_size)
        self.assertEqual(gm_reg_expected.xy_res, gm_reg_actual.xy_res)
        self.assertEqual(gm_reg_expected.xy_bbox, gm_reg_actual.xy_bbox)
        self.assertEqual(gm_reg_expected.crs, gm_reg_actual.crs)

    def test_2d_xy_coords(self):
        gm = GridMapping.from_coords(
            x_coords=xr.DataArray(
                [
                    [10.0, 10.1, 10.2, 10.3],
                    [10.1, 10.2, 10.3, 10.4],
                    [10.2, 10.3, 10.4, 10.5],
                ],
                dims=("lat", "lon"),
            ),
            y_coords=xr.DataArray(
                [
                    [52.0, 52.2, 52.4, 52.6],
                    [52.2, 52.4, 52.6, 52.8],
                    [52.4, 52.6, 52.8, 53.0],
                ],
                dims=("lat", "lon"),
            ),
            crs=GEO_CRS,
        )
        xy_coords = gm.xy_coords
        self.assertIsInstance(xy_coords, xr.DataArray)
        self.assertIs(xy_coords, gm.xy_coords)
        self.assertEqual(("coord", "lat", "lon"), xy_coords.dims)
        self.assertEqual((2, 3, 4), xy_coords.shape)
        self.assertEqual(("lon", "lat"), gm.xy_var_names)
        self.assertEqual(("lon", "lat"), gm.xy_dim_names)



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import math
from collections.abc import Iterable, Sequence

import dask
import dask.array as da
import numpy as np
import xarray as xr
from dask_image import ndinterp

from .constants import (
    AffineTransformMatrix,
    AggFunction,
    FillValues,
    FloatInt,
    PreventNaNPropagations,
    SpatialAggMethods,
    SpatialInterpMethodInt,
    SpatialInterpMethods,
)
from .gridmapping import GridMapping
from .utils import (
    _can_apply_affine_transform,
    _get_fill_value,
    _get_prevent_nan_propagation,
    _get_spatial_agg_method,
    _get_spatial_interp_method_int,
    _select_variables,
    normalize_grid_mapping,
)


def affine_transform_dataset(
    source_ds: xr.Dataset,
    target_gm: GridMapping,
    *,
    source_gm: GridMapping | None = None,
    variables: str | Iterable[str] | None = None,
    interp_methods: SpatialInterpMethods | None = None,
    agg_methods: SpatialAggMethods | None = None,
    prevent_nan_propagations: PreventNaNPropagations = False,
    fill_values: FillValues | None = None,
) -> xr.Dataset:
    """
    Apply an affine transformation to the spatial dimensions of a dataset,
    transforming it from the source grid mapping to the target grid mapping.

    Args:
        source_ds: The input dataset to be transformed.
        target_gm: The target grid mapping defining the spatial reference and
            output geometry.
        source_gm: The grid mapping of the input dataset. If None, it is inferred
            from the dataset.
        variables: Optional variable(s) to transform. If None, all variables are used.
        interp_methods: Optional interpolation method to be used for upsampling spatial
            data variables. Can be a single interpolation method for all variables or a
            dictionary mapping variable names or dtypes to interpolation method.
            Supported methods include:

            - `0` (nearest neighbor)
            - `1` (linear / bilinear)
            - `"nearest"`
            - `"bilinear"`

            The default is `0` for integer arrays, else `1`.
        agg_methods: Optional aggregation methods for downsampling spatial variables.
            Can be a single method for all variables or a dictionary mapping variable
            names or dtypes to methods. Supported methods include:
                "center", "count", "first", "last", "max", "mean", "median",
                "mode", "min", "prod", "std", "sum", and "var".
            Defaults to "center" for integer arrays, else "mean".
        prevent_nan_propagations: Optional boolean or mapping to prevent NaN
            propagation during upsampling (only applies when interpolation method
            is not nearest). Can be a single boolean or a dictionary mapping
            variable names or dtypes to booleans. Defaults to False.
        fill_values: Optional fill value(s) for areas outside the input bounds.
            Can be a single value or a dictionary mapping variable names or dtypes
            to fill values. If not provided, defaults are:

            - float: NaN
            - uint8: 255
            - uint16: 65535
            - other integers: -1

    Returns:
        A new dataset resampled and aligned to the target grid mapping.
            Data variables without spatial dimensions are copied to the output.
            Data variables with only one spatial dimension are ignored.
    """
    if source_gm is None:
        source_gm = GridMapping.from_dataset(source_ds)
    source_ds = normalize_grid_mapping(source_ds, source_gm)

    assert _can_apply_affine_transform(source_gm, target_gm), (
        f"Affine transformation cannot be applied to source CRS "
        f"{source_gm.crs.name!r} and target CRS {target_gm.crs.name!r}"
    )

    source_ds = _select_variables(source_ds, variables)

    target_ds = resample_dataset(
        source_ds,
        target_gm.ij_transform_to(source_gm),
        (source_gm.xy_dim_names[1], source_gm.xy_dim_names[0]),
        target_gm.size,
        target_gm.tile_size,
        interp_methods,
        agg_methods,
        prevent_nan_propagations,
        fill_values,
    )

    # assign coordinates from target grid-mapping
    x_name, y_name = target_gm.xy_var_names
    target_ds = target_ds.assign_coords(
        {x_name: target_gm.x_coords, y_name: target_gm.y_coords}
    )

    return target_ds


def resample_dataset(
    dataset: xr.Dataset,
    affine_matrix: AffineTransformMatrix,
    yx_dims: tuple[str, str],
    target_size: tuple[int, int],
    target_tile_size: tuple[int, int],
    interp_methods: SpatialInterpMethods | None = None,
    agg_methods: SpatialAggMethods | None = None,
    prevent_nan_propagations: PreventNaNPropagations = False,
    fill_values: FillValues | None = None,
) -> xr.Dataset:
    """
    Resample a dataset's spatial variables using an affine transformation.

    Applies resampling to 2D or 3D data arrays with spatial dimensions matching
    `yx_dims`. Variables that do not include the specified spatial dimensions
    are copied unchanged.

    Args:
        dataset: The input dataset containing spatial and non-spatial variables.
        affine_matrix: Affine transformation matrix mapping target to source coordinates.
        yx_dims: Tuple specifying the names of the spatial dimensions (y, x).
        target_size: The shape (height, width) of the resampled output in pixels.
        target_tile_size: Chunk size (height, width) for tiled output arrays.
        interp_methods: Optional interpolation method to be used for upsampling spatial
            data variables. Can be a single interpolation method for all variables or a
            dictionary mapping variable names or dtypes to interpolation method.
            Supported methods include:

            - `0` (nearest neighbor)
            - `1` (linear / bilinear)
            - `"nearest"`
            - `"bilinear"`

            The default is `0` for integer arrays, else `1`.
        agg_methods: Optional aggregation methods for downsampling spatial variables.
            Can be a single method for all variables or a dictionary mapping variable
            names or dtypes to methods. Supported methods include:
                "center", "count", "first", "last", "max", "mean", "median",
                "mode", "min", "prod", "std", "sum", and "var".
            Defaults to "center" for integer arrays, else "mean".
        prevent_nan_propagations: Optional boolean or mapping to prevent NaN
            propagation during upsampling (only applies when interpolation method
            is not nearest). Can be a single boolean or a dictionary mapping
            variable names or dtypes to booleans. Defaults to False.
        fill_values: Optional value(s) to use for regions outside source extent.
            Can be a single value or a dictionary mapping variable names or dtypes
            to specific fill values. If not provided, defaults are:

            - float: NaN
            - uint8: 255
            - uint16: 65535
            - other integers: -1

    Returns:
        A new dataset with spatial variables resampled to the target
            geometry. Non-spatial variables are preserved. Variables with only one
            spatial dimension are excluded.
    """
    data_vars = dict()
    coords = dict()
    for var_name, data_array in dataset.variables.items():
        data_array = xr.DataArray(data_array)
        new_data_array = None
        if data_array.dims[-2:] == yx_dims:
            assert len(data_array.dims) in (
                2,
                3,
            ), f"Data variable {var_name} has {len(data_array.dims)} dimensions."

            data_array_expanded = False
            if len(data_array.dims) == 2:
                data_array = data_array.expand_dims({"dummy": 1})
                data_array_expanded = True

            if isinstance(data_array.data, np.ndarray):
                is_numpy_array = True
                array = da.asarray(data_array.data)
            else:
                is_numpy_array = False
                array = data_array.data

            output_shape = array.shape[:-2] + (target_size[1], target_size[0])
            output_chunks = tuple(chunks[0] for chunks in array.chunks[:-2]) + (
                target_tile_size[1],
                target_tile_size[0],
            )
            resampled_array = _resample_array(
                array,
                affine_matrix,
                output_shape,
                output_chunks,
                _get_spatial_interp_method_int(interp_methods, var_name, data_array),
                _get_spatial_agg_method(agg_methods, var_name, data_array),
                _get_prevent_nan_propagation(
                    prevent_nan_propagations, var_name, data_array
                ),
                _get_fill_value(fill_values, var_name, data_array),
            )
            if is_numpy_array:
                resampled_array = resampled_array.compute()
            if data_array_expanded:
                resampled_array = resampled_array[0, :, :]
                dims = yx_dims
            else:
                dims = data_array.dims
            new_data_array = xr.DataArray(
                data=resampled_array, dims=dims, attrs=data_array.attrs
            )
        elif yx_dims[0] not in data_array.dims and yx_dims[1] not in data_array.dims:
            new_data_array = data_array
        if new_data_array is not None:
            if var_name in dataset.coords:
                coords[var_name] = new_data_array
            elif var_name in dataset.data_vars:
                data_vars[var_name] = new_data_array

    return xr.Dataset(data_vars=data_vars, coords=coords, attrs=dataset.attrs)


def _resample_array(
    array: da.Array,
    affine_matrix: AffineTransformMatrix,
    output_shape: Sequence[int],
    output_chunks: Sequence[int],
    interp_method: SpatialInterpMethodInt,
    agg_method: AggFunction,
    prevent_nan_propagation: bool,
    fill_value: FloatInt,
) -> da.Array:
    if (affine_matrix[0][0] > 1 or affine_matrix[1][1] > 1) and interp_method != 0:
        array = _downscale(
            array,
            affine_matrix,
            output_shape,
            output_chunks,
            agg_method,
            interp_method,
            prevent_nan_propagation,
            fill_value,
        )
    else:
        array = _upscale(
            array,
            affine_matrix,
            output_shape,
            output_chunks,
            interp_method,
            prevent_nan_propagation,
            fill_value,
        )
    return array


def _downscale(
    array: da.Array,
    affine_matrix: AffineTransformMatrix,
    output_shape: Sequence[int],
    output_chunks: Sequence[int],
    agg_method: AggFunction,
    interp_method: SpatialInterpMethodInt,
    prevent_nan_propagation: bool,
    fill_value: FloatInt,
) -> da.Array:
    (i_scale, _, i_off), (_, j_scale, j_off) = affine_matrix
    j_divisor = math.ceil(abs(j_scale))
    i_divisor = math.ceil(abs(i_scale))
    affine_matrix = (
        (i_scale / i_divisor, affine_matrix[0][1], affine_matrix[0][2]),
        (affine_matrix[1][0], j_scale / j_divisor, affine_matrix[1][2]),
    )
    output_shape = tuple(output_shape[:-2]) + (
        output_shape[-2] * j_divisor,
        output_shape[-1] * i_divisor,
    )

    array = _upscale(
        array,
        affine_matrix,
        output_shape,
        output_chunks,
        interp_method,
        prevent_nan_propagation,
        fill_value,
    )
    axes = {array.ndim - 2: j_divisor, array.ndim - 1: i_divisor}
    # noinspection PyTypeChecker
    array = da.coarsen(agg_method, array, axes)
    array = array.rechunk(output_chunks)

    return array


def _upscale(
    array: da.Array,
    affine_matrix: AffineTransformMatrix,
    output_shape: Sequence[int],
    output_chunks: Sequence[int],
    interp_method: SpatialInterpMethodInt,
    prevent_nan_propagation: bool,
    fill_value: FloatInt,
) -> da.Array:
    (i_scale, _, i_off), (_, j_scale, j_off) = affine_matrix
    offset = (j_off, i_off)
    matrix = np.diag((j_scale, i_scale))
    if interp_method > 1:
        raise ValueError(
            "interp_methods must be one of 0, 1, 'nearest', 'bilinear'. "
            "Higher order is not supported for 3D arrays in affine transforms, "
            "as it causes unintended blending across the non-spatial (e.g., time) "
            "dimension."
        )
    kwargs = dict(
        offset=offset,
        order=interp_method,
        output_shape=output_shape[-2:],
        output_chunks=output_chunks[-2:],
        mode="constant",
        cval=fill_value,
    )

    def _transform_slice(slice_2d: da.Array):
        if prevent_nan_propagation and interp_method > 0:
            # Prevent NaNs from spreading to nearby pixels during interpolation
            mask = da.isnan(slice_2d)
            # 1. replace NaN by zero
            filled_im = da.where(mask, 0.0, slice_2d)
            # 2. transform the zero-filled image
            scaled_im = ndinterp.affine_transform(filled_im, matrix, **kwargs)
            # 3. transform the inverted mask
            scaled_norm = ndinterp.affine_transform(1.0 - mask, matrix, **kwargs)
            # 4. put back NaN where there was zero,
            #    otherwise decode using scaled mask
            return da.where(
                da.isclose(scaled_norm, 0.0), np.nan, scaled_im / scaled_norm
            )
        else:
            return ndinterp.affine_transform(slice_2d, matrix, **kwargs)

    return da.stack([_transform_slice(array[i]) for i in range(array.shape[0])])



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import unittest

import numpy as np
import pyproj.crs

from xcube_resampling.gridmapping import CRS_CRS84, CRS_WGS84, GridMapping

from ..sampledata import create_s2plus_dataset

CRS_UTM_32N = pyproj.crs.CRS(32632)


# noinspection PyMethodMayBeStatic
class TransformTest(unittest.TestCase):
    def test_transform(self):
        gm = GridMapping.regular(
            size=(3, 3), xy_min=(10.05, 53.05), xy_res=0.1, crs=CRS_CRS84
        )
        gm_t = gm.transform(crs=CRS_UTM_32N)

        self.assertEqual(CRS_UTM_32N, gm_t.crs)
        self.assertEqual(False, gm_t.is_regular)
        self.assertEqual(("transformed_x", "transformed_y"), gm_t.xy_var_names)
        self.assertEqual(("lon", "lat"), gm_t.xy_dim_names)
        np.testing.assert_almost_equal(
            gm_t.xy_coords[0],
            np.array(
                [
                    [570057.076286, 576728.9360228, 583400.7295284],
                    [570220.3304187, 576907.7404859, 583595.0849538],
                    [570383.3684844, 577086.3083212, 583789.1831954],
                ]
            ),
        )
        np.testing.assert_almost_equal(
            gm_t.xy_coords[1],
            np.array(
                [
                    [5900595.928991, 5900698.5746648, 5900810.5532744],
                    [5889471.9033896, 5889574.6540572, 5889686.7472201],
                    [5878348.0594403, 5878450.9138481, 5878563.1201969],
                ]
            ),
        )

    def test_transform_xy_var_names(self):
        gm = GridMapping.regular(
            size=(3, 3), xy_min=(10, 53), xy_res=0.1, crs=CRS_CRS84
        )
        gm_t = gm.transform(crs=CRS_UTM_32N, xy_var_names=("x", "y"))
        self.assertEqual(CRS_UTM_32N, gm_t.crs)
        self.assertEqual(("x", "y"), gm_t.xy_var_names)
        self.assertEqual(("lon", "lat"), gm_t.xy_dim_names)

    def test_transform_s2(self):
        dataset = create_s2plus_dataset()

        gm = GridMapping.from_dataset(dataset, prefer_is_regular=True)
        # Assert we've picked the projected one which is regular
        self.assertIn("Projected", gm.crs.type_name)
        self.assertEqual(True, gm.is_regular)

        gm_t = gm.transform(CRS_CRS84)
        self.assertEqual(CRS_CRS84, gm_t.crs)

        gm_t = gm.transform(CRS_WGS84)
        self.assertEqual(CRS_WGS84, gm_t.crs)

    def test_transform_no_op(self):
        dataset = create_s2plus_dataset()

        gm = GridMapping.from_dataset(dataset, prefer_is_regular=True)
        # Assert we've picked the projected one which is regular
        self.assertIn("Projected", gm.crs.type_name)
        self.assertEqual(True, gm.is_regular)

        gm_t = gm.transform(gm.crs)
        self.assertIs(gm, gm_t)

        # Almost no op
        gm = GridMapping.regular(
            size=(3, 3), xy_min=(10, 53), xy_res=0.1, crs=CRS_CRS84
        )
        gm_t = gm.transform(crs=gm.crs, xy_var_names=("x", "y"))
        self.assertEqual(("x", "y"), gm_t.xy_var_names)



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import abc
import copy
import math
import threading
from collections.abc import Callable, Mapping, Sequence
from typing import Any

import numpy as np
import pyproj
import xarray as xr

from xcube_resampling.constants import AffineTransformMatrix, FloatInt

from .assertions import assert_given, assert_instance, assert_true
from .helpers import (
    _assert_valid_xy_coords,
    _assert_valid_xy_names,
    _from_affine,
    _normalize_int_pair,
    _normalize_number_pair,
    _to_affine,
    scale_xy_res_and_size,
)

# WGS84, axis order: lat, lon
CRS_WGS84 = pyproj.crs.CRS(4326)
# CRS84, axis order: lon, lat
CRS84 = "OGC:CRS84"
CRS_CRS84 = pyproj.crs.CRS.from_string(CRS84)

# Default tolerance for all operations that
# accept a key-word argument "tolerance":
DEFAULT_TOLERANCE = 1.0e-5


class GridMapping(abc.ABC):
    """An abstract base class for grid mappings that define an image grid and
    a transformation from image pixel coordinates to spatial Earth coordinates
    defined in a well-known coordinate reference system (CRS).

    This class cannot be instantiated directly. Use one of its factory methods
    to create instances:

    - `regular`
    - `regular_from_bbox`
    - `from_dataset`
    - `from_coords`

    Some instance methods can be used to derive new instances:

    - `derive`
    - `scale`
    - `transform`
    - `to_regular`

    Notes:

    - `x` : eastward axis in CRS coordinates
    - `y` : northward axis in CRS coordinates
    - `i` : eastward axis in pixel space
    - `j` : northward axis in pixel space

    References:
        [CF Conventions — Coordinate types](https://cfconventions.org/Data/cf-conventions/cf-conventions-1.13/cf-conventions.html#coordinate-types)
    """

    def __init__(
        self,
        /,
        size: int | tuple[int, int],
        tile_size: int | tuple[int, int] | None,
        xy_bbox: tuple[FloatInt, FloatInt, FloatInt, FloatInt],
        xy_res: FloatInt | tuple[FloatInt, FloatInt],
        crs: pyproj.crs.CRS,
        xy_var_names: tuple[str, str],
        xy_dim_names: tuple[str, str],
        is_regular: bool | None = None,
        is_lon_360: bool | None = None,
        is_j_axis_up: bool | None = None,
        x_coords: xr.DataArray | None = None,
        y_coords: xr.DataArray | None = None,
    ):
        width, height = _normalize_int_pair(size, name="size")
        assert_true(width > 1 and height > 1, "invalid size")

        tile_width, tile_height = _normalize_int_pair(
            tile_size, default=(width, height)
        )
        assert_true(tile_width > 1 and tile_height > 1, "invalid tile_size")

        assert_given(xy_bbox, name="xy_bbox")
        assert_given(xy_res, name="xy_res")
        _assert_valid_xy_names(xy_var_names, name="xy_var_names")
        _assert_valid_xy_names(xy_dim_names, name="xy_dim_names")
        assert_instance(crs, pyproj.crs.CRS, name="crs")

        if x_coords is not None:
            assert_instance(x_coords, xr.DataArray, name="x_coords")
            assert_true(
                x_coords.ndim in (1, 2),
                message=f"x_coords.ndim must be 1 or 2, was {x_coords.ndim}",
            )
        if y_coords is not None:
            assert_instance(y_coords, xr.DataArray, name="y_coords")
            assert_true(
                y_coords.ndim in (1, 2),
                message=f"y_coords.ndim must be 1 or 2, was {y_coords.ndim}",
            )

        x_res, y_res = _normalize_number_pair(xy_res, name="xy_res")
        assert_true(x_res > 0 and y_res > 0, "invalid xy_res")

        self._lock = threading.RLock()

        self._size = width, height
        self._tile_size = tile_width, tile_height
        self._xy_bbox = xy_bbox
        self._xy_res = x_res, y_res
        self._crs = crs
        self._xy_var_names = xy_var_names
        self._xy_dim_names = xy_dim_names
        self._is_regular = is_regular
        self._is_lon_360 = is_lon_360
        self._is_j_axis_up = is_j_axis_up
        self._x_coords = x_coords
        self._y_coords = y_coords
        self._xy_coords = None

    def derive(
        self,
        /,
        xy_var_names: tuple[str, str] = None,
        xy_dim_names: tuple[str, str] = None,
        tile_size: int | tuple[int, int] = None,
        is_j_axis_up: bool = None,
    ) -> "GridMapping":
        """Derive a new grid mapping from this one with some properties changed.

        Args:
            xy_var_names: The new x-, and y-variable names.
            xy_dim_names: The new x-, and y-dimension names.
            tile_size: The new tile size
            is_j_axis_up: Whether j-axis points up.

        Returns:
            A new, derived grid mapping.
        """
        other = copy.copy(self)
        if xy_var_names is not None:
            _assert_valid_xy_names(xy_var_names, name="xy_var_names")
            other._xy_var_names = xy_var_names
        if xy_dim_names is not None:
            _assert_valid_xy_names(xy_dim_names, name="xy_dim_names")
            other._xy_dim_names = xy_dim_names
        if tile_size is not None:
            tile_width, tile_height = _normalize_int_pair(tile_size, name="tile_size")
            assert_true(tile_width > 1 and tile_height > 1, "invalid tile_size")
            tile_size = tile_width, tile_height
            if other.tile_size != tile_size:
                other._tile_size = tile_width, tile_height
                with self._lock:
                    # if other._xy_coords has not been initialized before, we will do it
                    # in the next line. Otherwise, the following lines raise an error
                    if other._xy_coords is None:
                        _ = other.xy_coords
                    other._xy_coords = other._xy_coords.chunk(
                        {
                            dim: size
                            for (dim, size) in zip(
                                other._xy_coords.dims, other.xy_coords_chunks
                            )
                        }
                    )
        if is_j_axis_up is not None and is_j_axis_up != other._is_j_axis_up:
            other._is_j_axis_up = is_j_axis_up
            if other._y_coords is not None:
                other._y_coords = other._y_coords[::-1]
            if other._xy_coords is not None:
                other._xy_coords = other._xy_coords[:, ::-1, :]
                other._xy_coords = other._xy_coords.chunk(
                    {
                        dim: size
                        for (dim, size) in zip(
                            other._xy_coords.dims, other.xy_coords_chunks
                        )
                    }
                )

        return other

    def scale(
        self,
        xy_scale: FloatInt | tuple[FloatInt, FloatInt],
        tile_size: int | tuple[int, int] | None = None,
    ) -> "GridMapping":
        """Derive a scaled version of this regular grid mapping.

        Scaling factors larger than one correspond to up-scaling
        (pixels sizes decrease, image size increases).

        Scaling factors lower than one correspond to down-scaling.
        (pixels sizes increase, image size decreases).

        Args:
            xy_scale: The x-, and y-scaling factors. May be a single
                number or tuple.
            tile_size: The new tile size

        Returns:
            A new, scaled grid mapping.
        """
        self._assert_regular()
        x_scale, y_scale = _normalize_number_pair(xy_scale)
        new_xy_res, new_size = scale_xy_res_and_size(
            self.xy_res, self.size, (x_scale, y_scale)
        )
        if tile_size is not None:
            tile_width, tile_height = _normalize_int_pair(tile_size, name="tile_size")
        else:
            tile_width, tile_height = self.tile_size
        tile_width = min(new_size[0], tile_width)
        tile_height = min(new_size[1], tile_height)
        return self.regular(
            new_size,
            (self.xy_bbox[0] + new_xy_res[0] / 2, self.xy_bbox[1] + new_xy_res[1] / 2),
            new_xy_res,
            self.crs,
            tile_size=(tile_width, tile_height),
            is_j_axis_up=self.is_j_axis_up,
        ).derive(xy_dim_names=self.xy_dim_names, xy_var_names=self.xy_var_names)

    @property
    def size(self) -> tuple[int, int]:
        """Image size (width, height) in pixels."""
        return self._size

    @property
    def width(self) -> int:
        """Image width in pixels."""
        return self.size[0]

    @property
    def height(self) -> int:
        """Image height in pixels."""
        return self.size[1]

    @property
    def tile_size(self) -> tuple[int, int]:
        """Image tile size (width, height) in pixels."""
        return self._tile_size

    @property
    def is_tiled(self) -> bool:
        """Whether the image is tiled."""
        return self.size != self.tile_size

    @property
    def tile_width(self) -> int:
        """Image tile width in pixels."""
        return self.tile_size[0]

    @property
    def tile_height(self) -> int:
        """Image tile height in pixels."""
        return self.tile_size[1]

    @property
    def x_coords(self):
        """Return the eastward (x) coordinate values.

        The coordinates are provided as either a 1D or 2D array:
        - Shape (width,) for rectilinear grids
        - Shape (height, width) for curvilinear grids

        See Also:
            [CF Conventions, Longitude Coordinate](https://cfconventions.org/Data/cf-conventions/cf-conventions-1.13/cf-conventions.html#longitude-coordinate)
        """
        return self._get_computed_attribute("_x_coords", self._new_x_coords)

    @abc.abstractmethod
    def _new_x_coords(self) -> xr.DataArray:
        """Create new 1D or 2D x-coordinate array of
        shape (width,) or (height, width).
        """

    @property
    def y_coords(self):
        """Return the northward (y) coordinate values.

        The coordinates are provided as either a 1D or 2D array:
        - Shape (width,) for rectilinear grids
        - Shape (height, width) for curvilinear grids

        See Also:
            [CF Conventions, Latitude Coordinate](https://cfconventions.org/Data/cf-conventions/cf-conventions-1.13/cf-conventions.html#latitude-coordinate)
        """
        return self._get_computed_attribute("_y_coords", self._new_y_coords)

    @abc.abstractmethod
    def _new_y_coords(self) -> xr.DataArray:
        """Create new 1D or 2D y-coordinate array of
        shape (width,) or (height, width).
        """

    @property
    def xy_coords(self) -> xr.DataArray:
        """Return the spatial coordinates as a single DataArray.

        The coordinates are provided as a 3D array of shape `(2, height, width)`,
        where the first axis corresponds to the x and y coordinates stacked in order:
        `(x_coords, y_coords)`. All values are expressed in the units of the
        dataset's CRS (coordinate reference system).
        """
        xy_coords = self._get_computed_attribute("_xy_coords", self._new_xy_coords)
        _assert_valid_xy_coords(xy_coords)
        return xy_coords

    @property
    def xy_coords_chunks(self) -> tuple[int, int, int]:
        """Return the chunks for the `xy_coords` array."""
        return 2, self.tile_height, self.tile_width

    @abc.abstractmethod
    def _new_xy_coords(self) -> xr.DataArray:
        """Create new coordinate array of shape (2, height, width)."""

    def _get_computed_attribute(self, name: str, computer: Callable[[], Any]) -> Any:
        """Get the value for a computed attribute.
        Utility to be used by this and derived classes.
        """
        value = getattr(self, name)
        if value is not None:
            return value
        with self._lock:
            # Double null check
            value = getattr(self, name)
            if value is not None:
                return value
            value = computer()
            setattr(self, name, value)
            return value

    @property
    def xy_var_names(self) -> tuple[str, str]:
        """The variable names of `x_coords` and `y_coords` as tuple
        `(x_var_name, y_var_name)`.
        """
        return self._xy_var_names

    @property
    def xy_dim_names(self) -> tuple[str, str]:
        """The dimension names of `x_coords` and `y_coords` as
        tuple `(x_dim_name, y_dim_name)`.
        """
        return self._xy_dim_names

    @property
    def xy_bbox(self) -> tuple[float, float, float, float]:
        """The image's bounding box in CRS coordinates."""
        return self._xy_bbox

    @property
    def x_min(self) -> FloatInt:
        """Minimum eastward (x) coordinate value in CRS units."""
        return round(self._xy_bbox[0] + self.x_res / 2, ndigits=7)

    @property
    def y_min(self) -> FloatInt:
        """Minimum northward (y) coordinate value in CRS units."""
        return round(self._xy_bbox[1] + self.y_res / 2, ndigits=7)

    @property
    def x_max(self) -> FloatInt:
        """Maximum eastward (x) coordinate value in CRS units."""
        return round(self._xy_bbox[2] - self.x_res / 2, ndigits=7)

    @property
    def y_max(self) -> FloatInt:
        """Maximum northward (y) coordinate value in CRS units."""
        return round(self._xy_bbox[3] - self.y_res / 2, ndigits=7)

    @property
    def xy_res(self) -> tuple[FloatInt, FloatInt]:
        """Pixel size in eastward (x) and northward (y) direction `(x_res, y_res)`."""
        return self._xy_res

    @property
    def x_res(self) -> FloatInt:
        """Pixel size in CRS units per pixel in eastward (x) direction."""
        return self._xy_res[0]

    @property
    def y_res(self) -> FloatInt:
        """Pixel size in CRS units per pixel in northward (y) direction."""
        return self._xy_res[1]

    @property
    def crs(self) -> pyproj.crs.CRS:
        """The coordinate reference system."""
        return self._crs

    @property
    def spatial_unit_name(self) -> str:
        return self._crs.axis_info[0].unit_name

    @property
    def is_lon_360(self) -> bool | None:
        """Check whether `x_max` is greater than 180 degrees.
        Effectively tests whether the range `x_min`, `x_max` crosses
        the anti-meridian at 180 degrees.
        Works only for geographical coordinate reference systems.
        """
        return self._is_lon_360

    @property
    def is_regular(self) -> bool | None:
        """Check whether the x (eastward) and y (northward) coordinates form a regular
        grid.

        A regular grid is defined as having constant spacing (delta) along
        the x- and y-axes individually.
        """

        return self._is_regular

    @property
    def is_j_axis_up(self) -> bool | None:
        """Indicate whether the positive image j-axis points upward.

        Here, the j-axis refers to the y-coordinate in pixel space. By convention,
        the positive j-axis usually points downward in image coordinates, so
        `True` indicates that the axis has been flipped to point upward.
        """
        return self._is_j_axis_up

    @property
    def ij_to_xy_transform(self) -> AffineTransformMatrix:
        """The affine transformation matrix from image to CRS coordinates.
        Defined only for grid mappings with regular x (eastward) and y (northward)
        coordinates.
        """
        self._assert_regular()
        if self.is_j_axis_up:
            return (
                (self.x_res, 0.0, self.x_min),
                (0.0, self.y_res, self.y_min),
            )
        else:
            return (
                (self.x_res, 0.0, self.x_min),
                (0.0, -self.y_res, self.y_max),
            )

    @property
    def xy_to_ij_transform(self) -> AffineTransformMatrix:
        """The affine transformation matrix from CRS to image coordinates.
        Defined only for grid mappings with regular x (eastward) and y (northward)
        coordinates.
        """
        self._assert_regular()
        return _from_affine(~_to_affine(self.ij_to_xy_transform))

    def ij_transform_to(self, other: "GridMapping") -> AffineTransformMatrix:
        """Get the affine transformation matrix that transforms
        image coordinates of `other` grid-mapping into image coordinates
        of this image geometry.

        Defined only for grid mappings with regular x (eastward) and y (northward)
        coordinates.

        Args:
            other: The other grid-mapping

        Returns:
            Affine transformation matrix
        """
        self._assert_regular()
        self.assert_regular(other, name="other")
        a = _to_affine(self.ij_to_xy_transform)
        b = _to_affine(other.xy_to_ij_transform)
        return _from_affine(b * a)

    def ij_transform_from(self, other: "GridMapping") -> AffineTransformMatrix:
        """Get the affine transformation matrix that transforms
        image coordinates of this grid-mapping to image
        coordinates of *other* grid-mapping.

        Defined only for grid mappings with regular x (eastward) and y (northward)
        coordinates.

        Args:
            other: The other grid-mapping

        Returns:
            Affine transformation matrix
        """
        self._assert_regular()
        self.assert_regular(other, name="other")
        a = _to_affine(self.ij_transform_to(other))
        return _from_affine(~a)

    @property
    def ij_bbox(self) -> tuple[int, int, int, int]:
        """The image's bounding box in pixel coordinates."""
        return 0, 0, self.width, self.height

    @property
    def ij_bboxes(self) -> np.ndarray:
        """The image tiles' bounding boxes in image pixel coordinates."""
        h, w = self.height, self.width
        th, tw = self.tile_height, self.tile_width

        y_starts = range(0, h, th)
        x_starts = range(0, w, tw)

        n = len(y_starts) * len(x_starts)
        ij_bboxes = np.empty((n, 4), dtype=np.int64)

        k = 0
        for y0 in y_starts:
            y1 = min(y0 + th, h)
            for x0 in x_starts:
                x1 = min(x0 + tw, w)
                ij_bboxes[k] = (x0, y0, x1, y1)
                k += 1

        return ij_bboxes

    @property
    def xy_bboxes(self) -> np.ndarray:
        """The image tiles' bounding boxes in CRS coordinates."""
        x_pad, y_pad = self.x_res / 2, self.y_res / 2
        if self.is_j_axis_up:
            xy_offset = np.array(
                [
                    self.x_min - x_pad,
                    self.y_min - y_pad,
                    self.x_min - x_pad,
                    self.y_min - y_pad,
                ]
            )
            xy_scale = np.array([self.x_res, self.y_res, self.x_res, self.y_res])
            xy_bboxes = xy_offset + xy_scale * self.ij_bboxes
        else:
            xy_offset = np.array(
                [
                    self.x_min - x_pad,
                    self.y_max + y_pad,
                    self.x_min - x_pad,
                    self.y_max + y_pad,
                ]
            )
            xy_scale = np.array([self.x_res, -self.y_res, self.x_res, -self.y_res])
            xy_bboxes = xy_offset + xy_scale * self.ij_bboxes
            xy_bboxes[:, [1, 3]] = xy_bboxes[:, [3, 1]]
        return xy_bboxes

    def to_coords(
        self,
        xy_var_names: tuple[str, str] = None,
        xy_dim_names: tuple[str, str] = None,
        exclude_bounds: bool = False,
        reuse_coords: bool = False,
    ) -> Mapping[str, xr.DataArray]:
        """Get CF-compliant axis coordinate variables and cell boundary
        coordinate variables.

        Defined only for grid mappings with regular x,y coordinates.

        Args:
            xy_var_names: Optional coordinate variable names
                (x_var_name, y_var_name).
            xy_dim_names: Optional coordinate dimensions names
                (x_dim_name, y_dim_name).
            exclude_bounds: If True, do not create bounds coordinates.
                Defaults to False.
            reuse_coords: Whether to either reuse target coordinate
                arrays from target_gm or to compute new ones.

        Returns:
            dictionary with coordinate variables
        """
        self._assert_regular()
        from .coords import grid_mapping_to_coords

        return grid_mapping_to_coords(
            self,
            xy_var_names=xy_var_names,
            xy_dim_names=xy_dim_names,
            exclude_bounds=exclude_bounds,
            reuse_coords=reuse_coords,
        )

    def transform(
        self,
        crs: str | pyproj.crs.CRS,
        *,
        tile_size: int | tuple[int, int] = None,
        xy_var_names: tuple[str, str] = None,
        tolerance: float = DEFAULT_TOLERANCE,
    ) -> "GridMapping":
        """Transform this grid mapping so it uses the given
        spatial coordinate reference system into another *crs*.

        Args:
            crs: The new spatial coordinate reference system.
            tile_size: Optional new tile size.
            xy_var_names: Optional new coordinate names.
            tolerance: Absolute tolerance used when comparing
                coordinates with each other. Must be in the units of the
                *crs* and must be greater zero.

        Returns:
            A new grid mapping that uses *crs*.
        """
        from .transform import transform_grid_mapping

        return transform_grid_mapping(
            self,
            crs,
            tile_size=tile_size,
            xy_var_names=xy_var_names,
            tolerance=tolerance,
        )

    @classmethod
    def regular(
        cls,
        size: int | tuple[int, int],
        xy_min: tuple[float, float],
        xy_res: float | tuple[float, float],
        crs: str | pyproj.crs.CRS,
        *,
        tile_size: int | tuple[int, int] = None,
        is_j_axis_up: bool = False,
    ) -> "GridMapping":
        """Create a new regular grid mapping.

        Args:
            size: Size in pixels.
            xy_min: Minimum x- and y-coordinates.
            xy_res: Resolution in x- and y-directions.
            crs: Spatial coordinate reference system.
            tile_size: Optional tile size.
            is_j_axis_up: Whether positive j-axis points up. Defaults to
                false.

        Returns:
            A new regular grid mapping.
        """
        from .regular import new_regular_grid_mapping

        return new_regular_grid_mapping(
            size=size,
            xy_min=xy_min,
            xy_res=xy_res,
            crs=crs,
            tile_size=tile_size,
            is_j_axis_up=is_j_axis_up,
        )

    @classmethod
    def regular_from_bbox(
        cls,
        bbox: Sequence[FloatInt],
        xy_res: float | tuple[float, float],
        crs: str | pyproj.crs.CRS,
        tile_size: int | tuple[int, int] = None,
        is_j_axis_up: bool = False,
    ) -> "GridMapping":
        """Creates a regular grid mapping for a given coordinate reference system based
        on the given bounding box and spatial resolution.

        Args:
            bbox: Bounding box coordinates in the format [west, south, east, north].
                The values must be in the given CRS.
            xy_res: Resolution in x- and y-directions.
            crs: Coordinate reference system (e.g., "EPSG:4326").
            tile_size: Chunk size for the grid; if a single int is given,
                square chunk size is assumed. Defaults to 1024.
            is_j_axis_up: Whether positive j-axis points up. Defaults to False.

        Returns:
            A regular grid mapping object.
        """
        from .regular import new_regular_grid_mapping_from_bbox

        return new_regular_grid_mapping_from_bbox(
            bbox, xy_res, crs, tile_size=tile_size, is_j_axis_up=is_j_axis_up
        )

    def to_regular(
        self, tile_size: int | tuple[int, int] | None = None, is_j_axis_up: bool = False
    ) -> "GridMapping":
        """Transform this grid mapping into one that is regular.

        Args:
            tile_size: Optional tile size.
            is_j_axis_up: Whether positive j-axis points up. Defaults to
                false.

        Returns:
            A new regular grid mapping or this grid mapping, if it is
            already regular.
        """
        from .regular import to_regular_grid_mapping

        return to_regular_grid_mapping(
            self, tile_size=tile_size, is_j_axis_up=is_j_axis_up
        )

    @classmethod
    def from_dataset(
        cls,
        dataset: xr.Dataset,
        *,
        crs: str | pyproj.crs.CRS | None = None,
        tile_size: int | tuple[int, int] | None = None,
        prefer_is_regular: bool = True,
        prefer_crs: str | pyproj.crs.CRS | None = None,
        emit_warnings: bool = False,
        tolerance: float = DEFAULT_TOLERANCE,
    ) -> "GridMapping":
        """Create a grid mapping for the given *dataset*.

        Args:
            dataset: The dataset.
            crs: Optional spatial coordinate reference system.
            tile_size: Optional tile size
            prefer_is_regular: Whether to prefer a regular grid mapping
                if multiple found. Default is True.
            prefer_crs: The preferred CRS of a grid mapping if multiple
                found.
            emit_warnings: Whether to emit warning for non-CF compliant
                datasets.
            tolerance: Absolute tolerance used when comparing
                coordinates with each other. Must be in the units of the
                *crs* and must be greater zero.

        Returns:
            a new grid mapping instance.
        """
        from .dataset import new_grid_mapping_from_dataset

        return new_grid_mapping_from_dataset(
            dataset=dataset,
            crs=crs,
            tile_size=tile_size,
            prefer_is_regular=prefer_is_regular,
            prefer_crs=prefer_crs,
            emit_warnings=emit_warnings,
            tolerance=tolerance,
        )

    @classmethod
    def from_coords(
        cls,
        x_coords: xr.DataArray,
        y_coords: xr.DataArray,
        crs: str | pyproj.crs.CRS,
        *,
        tile_size: int | tuple[int, int] | None = None,
        tolerance: float = DEFAULT_TOLERANCE,
    ) -> "GridMapping":
        """Create a grid mapping from given x- and y-coordinates
        *x_coords*, *y_coords* and spatial coordinate reference
        system *crs*.

        Args:
            x_coords: The x-coordinates.
            y_coords: The y-coordinates.
            crs: The spatial coordinate reference system.
            tile_size: Optional tile size.
            tolerance: Absolute tolerance used when comparing
                coordinates with each other. Must be in the units of the
                *crs* and must be greater zero.

        Returns:
            A new grid mapping.
        """
        from .coords import new_grid_mapping_from_coords

        return new_grid_mapping_from_coords(
            x_coords=x_coords,
            y_coords=y_coords,
            crs=crs,
            tile_size=tile_size,
            tolerance=tolerance,
        )

    def is_close(
        self, other: "GridMapping", tolerance: float = DEFAULT_TOLERANCE
    ) -> bool:
        """Tests whether this grid mapping is close to *other*.

        Args:
            other: The other grid mapping.
            tolerance: Absolute tolerance used when comparing
                coordinates with each other. Must be in the units of the
                *crs* and must be greater zero.

        Returns:
            True, if so, False otherwise.
        """
        if self is other:
            return True
        if (
            self.is_j_axis_up == other.is_j_axis_up
            and self.is_lon_360 == other.is_lon_360
            and self.is_regular == other.is_regular
            and self.size == other.size
            and self.tile_size == other.tile_size
            and self.crs == other.crs
        ):
            sxr, syr = self.xy_res
            oxr, oyr = other.xy_res
            if math.isclose(sxr, oxr, abs_tol=tolerance) and math.isclose(
                syr, oyr, abs_tol=tolerance
            ):
                sx1, sy1, sx2, sy2 = self.xy_bbox
                ox1, oy1, ox2, oy2 = other.xy_bbox
                return (
                    math.isclose(sx1, ox1, abs_tol=tolerance)
                    and math.isclose(sy1, oy1, abs_tol=tolerance)
                    and math.isclose(sx2, ox2, abs_tol=tolerance)
                    and math.isclose(sy2, oy2, abs_tol=tolerance)
                )
        return False

    @classmethod
    def assert_regular(cls, value: Any, name: str = None):
        assert_instance(value, GridMapping, name=name)
        if not value.is_regular:
            raise ValueError(f"{name or 'value'} must be a regular grid mapping")

    def _assert_regular(self):
        if not self.is_regular:
            raise NotImplementedError(
                "Operation not implemented for non-regular grid mappings"
            )

    def _repr_markdown_(self) -> str:
        """Generate an IPython Notebook Markdown representation."""
        is_regular = self.is_regular if self.is_regular is not None else "_unknown_"
        is_j_axis_up = (
            self.is_j_axis_up if self.is_j_axis_up is not None else "_unknown_"
        )
        is_lon_360 = self.is_lon_360 if self.is_lon_360 is not None else "_unknown_"
        xy_res = repr(self.xy_res) + ("" if self.is_regular else "  _estimated_")
        return "\n".join(
            [
                f"class: **{self.__class__.__name__}**",
                f"* is_regular: {is_regular}",
                f"* is_j_axis_up: {is_j_axis_up}",
                f"* is_lon_360: {is_lon_360}",
                f"* crs: {self.crs}",
                f"* xy_res: {xy_res}",
                f"* xy_bbox: {self.xy_bbox}",
                f"* ij_bbox: {self.ij_bbox}",
                f"* xy_dim_names: {self.xy_dim_names}",
                f"* xy_var_names: {self.xy_var_names}",
                f"* size: {self.size}",
                f"* tile_size: {self.tile_size}",
            ]
        )



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import os
from collections.abc import Hashable, Iterable, Mapping

import dask.array as da
import numba as nb
import numpy as np
import xarray as xr

from .affine import resample_dataset
from .constants import (
    SCALE_LIMIT,
    FillValues,
    FloatInt,
    PreventNaNPropagations,
    SpatialAggMethods,
    SpatialInterpMethod,
    SpatialInterpMethods,
    SpatialInterpMethodStr,
)
from .gridmapping import GridMapping
from .gridmapping.helpers import from_lon_360
from .utils import (
    SourceTileIndexing,
    _clip_if_needed,
    _get_fill_value,
    _get_spatial_interp_method_str,
    _get_spatial_interp_methods,
    _is_equal_crs,
    _map_to_source_indices,
    _prep_spatial_interp_methods_downscale,
    _reorganize_tiled_array,
    _sample_array_at_indices,
    _select_variables,
    bbox_overlap,
    normalize_grid_mapping,
)

FX_MIN = FY_MIN = 1e-3
FXFY_MAX = 1.0 + 2 * 1e-3


def rectify_dataset(
    source_ds: xr.Dataset,
    *,
    target_gm: GridMapping | None = None,
    source_gm: GridMapping | None = None,
    variables: str | Iterable[str] | None = None,
    interp_methods: SpatialInterpMethods | None = None,
    agg_methods: SpatialAggMethods | None = None,
    prevent_nan_propagations: PreventNaNPropagations = False,
    fill_values: FillValues | None = None,
    tile_size: int | tuple[int, int] | None = None,
    output_indices_names: tuple[str, str] | None = None,
) -> xr.Dataset:
    """
    Rectify a dataset with non-regular grid to a regular grid defined by a target
    grid mapping.

    This function transforms spatial coordinates to a regular grid while preserving
    data values. It optionally downsamples high-resolution inputs prior to rectifying.

    Args:
        source_ds: The source dataset with 2D spatial coordinate variables.
        target_gm: Optional target grid mapping defining the output regular grid.
            If not provided, one is derived from the source grid mapping.
        source_gm: Optional grid mapping of the source dataset. If not given, it is
            inferred from the dataset.
        variables: Optional variable(s) to rectify. If None, all eligible variables
            are processed.
        interp_methods: Optional interpolation method to be used for upsampling spatial
            data variables. Can be a single interpolation method for all variables or a
            dictionary mapping variable names or dtypes to interpolation method.
            Supported methods include:

            - `0` (nearest neighbor)
            - `1` (linear / bilinear)
            - `"nearest"`
            - `"triangular"`
            - `"bilinear"`

            The default is `0` for integer arrays, else `1`.
        agg_methods: Optional aggregation methods for downsampling spatial variables.
            Can be a single method for all variables or a dictionary mapping variable
            names or dtypes to methods. Supported methods include:
                "center", "count", "first", "last", "max", "mean", "median",
                "mode", "min", "prod", "std", "sum", and "var".
            Defaults to "center" for integer arrays, else "mean".
        prevent_nan_propagations: Optional boolean or mapping to prevent NaN
            propagation during upsampling (only applies when interpolation method
            is not nearest). Can be a single boolean or a dictionary mapping
            variable names or dtypes to booleans. Defaults to False.
        fill_values: Optional fill value(s) for areas outside input coverage.
            Can be a single value or dictionary by variable or type. If not provided,
            defaults based on data type are used:

            - float: NaN
            - uint8: 255
            - uint16: 65535
            - other ints: -1

        tile_size: Optional tile size for inferring a regular grid, if `target_gm` is
            not provided.
        output_indices_names: Optional names for two variables that store the source
            pixel indices for the last and second-last dimension, respectively.

    Returns:
        A new dataset with spatial variables rectified to a regular grid.
        Variables not having 2D spatial dimensions are copied as-is. 1D spatial
        coordinate variables are ignored in the output.
    """
    source_ds = _select_variables(source_ds, variables)

    source_gm = source_gm or GridMapping.from_dataset(source_ds)
    source_ds = normalize_grid_mapping(source_ds, source_gm)

    target_gm = target_gm or source_gm.to_regular(tile_size=tile_size)

    source_ds, source_gm = _align_source_crs_to_target(source_ds, source_gm, target_gm)
    source_ds, source_gm, is_empty = _clip_if_needed(
        source_ds, source_gm, target_gm, fill_values
    )
    if is_empty:
        return source_ds

    # If source has higher resolution than target, downscale first, then reproject
    interp_methods = _get_spatial_interp_methods(source_ds, interp_methods)
    source_ds, source_gm = _downscale_source_dataset(
        source_ds,
        source_gm,
        target_gm,
        interp_methods,
        agg_methods,
        prevent_nan_propagations,
    )

    # get fractional location each target pixel in source grid mapping
    indexing = _compute_source_tile_indexing(source_gm, target_gm)
    src_x = _reorganize_tiled_array(
        da.asarray(source_gm.x_coords.data),
        indexing,
        _get_fill_value(
            fill_values, source_gm.xy_var_names[0], source_ds[source_gm.xy_var_names[0]]
        ),
    )
    src_y = _reorganize_tiled_array(
        da.asarray(source_gm.y_coords.data),
        indexing,
        _get_fill_value(
            fill_values, source_gm.xy_var_names[1], source_ds[source_gm.xy_var_names[1]]
        ),
    )
    pixel_target_ij = _get_pixel_target_ij(src_x, src_y, target_gm)

    return _assemble_rectified_dataset(
        source_ds,
        source_gm,
        target_gm,
        indexing,
        pixel_target_ij,
        interp_methods,
        fill_values,
        output_indices_names,
    )


def _assemble_rectified_dataset(
    source_ds: xr.Dataset,
    source_gm: GridMapping,
    target_gm: GridMapping,
    indexing: SourceTileIndexing,
    pixel_target_ij: da.Array,
    interp_methods: SpatialInterpMethods | None,
    fill_values: FillValues | None,
    output_indices_names: tuple[str, str] | None,
) -> xr.Dataset:
    """
    Assemble the final rectified dataset on the target grid.
    """
    src_x_name, src_y_name = source_gm.xy_var_names
    tgt_x_name, tgt_y_name = target_gm.xy_var_names

    coords = source_ds.coords.to_dataset()
    coords = coords.drop_vars((src_x_name, src_y_name), errors="ignore")

    tgt_x = target_gm.x_coords
    if target_gm.is_lon_360:
        tgt_x = from_lon_360(tgt_x)
    coords[tgt_x_name] = tgt_x
    coords[tgt_y_name] = target_gm.y_coords
    coords["spatial_ref"] = xr.DataArray(0, attrs=target_gm.crs.to_cf())
    target_ds = xr.Dataset(coords=coords, attrs=source_ds.attrs)

    yx_dims = (source_gm.xy_dim_names[1], source_gm.xy_dim_names[0])
    for var_name, da_src in source_ds.data_vars.items():
        if da_src.dims[-2:] == yx_dims:
            assert len(da_src.dims) in (
                2,
                3,
            ), f"Data variable {var_name} has {len(da_src.dims)} dimensions."
            fill_value = _get_fill_value(fill_values, var_name, da_src)
            interp_method = _get_spatial_interp_method_str(
                interp_methods, var_name, da_src
            )

            rectified = _rectify_data_array(
                da_src.data,
                indexing,
                pixel_target_ij,
                target_gm,
                interp_method,
                fill_value,
            )
            dims = da_src.dims[:-2] + (
                target_gm.xy_dim_names[1],
                target_gm.xy_dim_names[0],
            )
            target_ds[var_name] = xr.DataArray(rectified, dims=dims, attrs=da_src.attrs)

        elif yx_dims[0] not in da_src.dims and yx_dims[1] not in da_src.dims:
            # Non-spatial variable → copy as-is
            target_ds[var_name] = da_src

    # optional source pixel index outputs
    if output_indices_names:
        output_indices = _map_to_source_indices(pixel_target_ij, indexing)
        target_ds[output_indices_names[0]] = (
            (tgt_y_name, tgt_x_name),
            output_indices[0],
        )
        target_ds[output_indices_names[1]] = (
            (tgt_y_name, tgt_x_name),
            output_indices[1],
        )

    return target_ds


def _align_source_crs_to_target(
    source_ds: xr.Dataset,
    source_gm: GridMapping,
    target_gm: GridMapping,
) -> tuple[xr.Dataset, GridMapping]:
    """
    Ensure source dataset coordinates are in the same CRS as the target grid mapping.
    """
    if _is_equal_crs(source_gm, target_gm):
        return source_ds, source_gm

    source_gm_transformed = source_gm.transform(target_gm.crs)
    source_ds = source_ds.drop_vars(source_gm.xy_var_names)
    source_ds = source_ds.assign_coords(
        {
            "spatial_ref": xr.DataArray(0, attrs=target_gm.crs.to_cf()),
            "transformed_x": source_gm_transformed.x_coords,
            "transformed_y": source_gm_transformed.y_coords,
        }
    )
    source_gm = GridMapping.from_dataset(source_ds)
    return source_ds, source_gm


def _downscale_source_dataset(
    source_ds: xr.Dataset,
    source_gm: GridMapping,
    target_gm: GridMapping,
    interp_methods: Mapping[Hashable, SpatialInterpMethod],
    agg_methods: SpatialAggMethods | None,
    prevent_nan_propagations: PreventNaNPropagations,
) -> (xr.Dataset, GridMapping):
    if all(v in (0, "nearest") for v in interp_methods.values()):
        return source_ds, source_gm

    x_scale = source_gm.x_res / target_gm.x_res
    y_scale = source_gm.y_res / target_gm.y_res
    if x_scale >= SCALE_LIMIT and y_scale >= SCALE_LIMIT:
        return source_ds, source_gm

    w = np.floor(x_scale * (source_gm.width - 1))
    h = np.floor(y_scale * (source_gm.height - 1))
    downscaled_size = (w if w >= 2 else 2, h if h >= 2 else 2)
    source_ds = resample_dataset(
        source_ds,
        ((1 / x_scale, 0, 0), (0, 1 / y_scale, 0)),
        (source_gm.xy_dim_names[1], source_gm.xy_dim_names[0]),
        downscaled_size,
        source_gm.tile_size,
        _prep_spatial_interp_methods_downscale(interp_methods),
        agg_methods,
        prevent_nan_propagations,
    )
    source_gm = GridMapping.from_dataset(source_ds)

    return source_ds, source_gm


def _get_pixel_target_ij(
    src_x_coords: da.Array,
    src_y_coords: da.Array,
    target_gm: GridMapping,
) -> da.Array:
    """
    Compute fractional source pixel indices for each target grid pixel.

    For each target pixel, this function determines the fractional (i, j)
    location in the source grid using forward mapping from source grid
    quadrilaterals to target grid pixels.

    The computation is performed blockwise using Dask. Each block receives:
    - a 1D slice of target x coordinates
    - a 1D slice of target y coordinates
    - a 2D tile of source x/y coordinates

    Args:
        src_x_coords: Source grid x-coordinates (2D, tiled).
        src_y_coords: Source grid y-coordinates (2D, tiled).
        target_gm: Target grid mapping.

    Returns:
        Dask array of shape (2, target_height, target_width) containing
        fractional source pixel indices (ix, iy) per target pixel.
        Missing values are NaN.
    """
    dtype = src_x_coords.dtype

    def _pixel_index_block(
        tgt_x_block: np.ndarray,
        tgt_y_block: np.ndarray,
        src_x_block: np.ndarray,
        src_y_block: np.ndarray,
        scale: np.ndarray,
    ) -> np.ndarray:
        """Compute fractional source indices for a single target tile."""
        target_x = tgt_x_block[0, :]
        target_y = tgt_y_block[:, 0]

        nx_tgt = target_x.size
        ny_tgt = target_y.size

        idx_frac = np.full(
            (2, ny_tgt, nx_tgt),
            np.nan,
            dtype=np.float32 if np.issubdtype(dtype, np.integer) else dtype,
        )

        ny_src, nx_src = src_x_block.shape

        # Quad origin indices (top-left corner per quad)
        ix0 = np.repeat(np.arange(nx_src - 1), ny_src - 1)
        iy0 = np.tile(np.arange(ny_src - 1), nx_src - 1)
        quad_ij0 = np.stack((ix0, iy0)).astype(np.int32)

        # Quad corner coordinates
        src_coords = np.stack((src_x_block, src_y_block))
        quad_corners = np.stack(
            (
                src_coords[:, quad_ij0[1], quad_ij0[0]],
                src_coords[:, quad_ij0[1], quad_ij0[0] + 1],
                src_coords[:, quad_ij0[1] + 1, quad_ij0[0]],
                src_coords[:, quad_ij0[1] + 1, quad_ij0[0] + 1],
            ),
            axis=1,
        )

        valid = ~np.any(np.isnan(quad_corners[0]), axis=0)
        quad_corners = quad_corners[:, :, valid]
        quad_ij0 = quad_ij0[:, valid]

        if quad_corners.size == 0:
            return idx_frac

        # Bounding target pixel indices per quad
        offset = np.array([target_x[0], target_y[0]])
        tgt_bbox = np.floor(
            (quad_corners - offset[:, None, None]) / scale[:, None, None]
        ).astype(np.int32)
        tgt_bbox = np.vstack((tgt_bbox.min(axis=1), tgt_bbox.max(axis=1)))

        valid = (
            (tgt_bbox[2] >= 0)
            & (tgt_bbox[3] >= 0)
            & (tgt_bbox[0] < nx_tgt)
            & (tgt_bbox[1] < ny_tgt)
        )

        tgt_bbox = tgt_bbox[:, valid]
        quad_corners = quad_corners[:, :, valid]
        quad_ij0 = quad_ij0[:, valid]

        if tgt_bbox.size == 0:
            return idx_frac

        tgt_bbox[0:2] = np.maximum(tgt_bbox[0:2], 0)
        tgt_bbox[2] = np.minimum(tgt_bbox[2], nx_tgt - 1)
        tgt_bbox[3] = np.minimum(tgt_bbox[3], ny_tgt - 1)

        _fill_pixel_indices_from_quads(
            quad_corners,
            tgt_bbox,
            idx_frac,
            target_x,
            target_y,
            quad_ij0,
        )

        return idx_frac

    # Build block-aligned target coordinate views
    tgt_x = target_gm.x_coords.data
    tgt_y = target_gm.y_coords.data
    scale = np.array([tgt_x[1] - tgt_x[0], tgt_y[1] - tgt_y[0]])
    tgt_x = da.stack([tgt_x] * len(tgt_y.chunks[0]), axis=0)
    tgt_y = da.stack([tgt_y] * len(tgt_x.chunks[1]), axis=1)

    return da.map_blocks(
        _pixel_index_block,
        tgt_x,
        tgt_y,
        src_x_coords,
        src_y_coords,
        scale,
        dtype=np.float32 if np.issubdtype(dtype, np.integer) else dtype,
        chunks=(2, tgt_y.chunks[0], tgt_x.chunks[1]),
    )


def _env_flag(name: str) -> bool:
    return os.environ.get(name, "").lower() in ("1", "true", "yes")


@nb.njit(
    nogil=True,
    cache=not _env_flag("XCUBE_NUMBA_CACHE_DISABLE"),
)
def _fill_pixel_indices_from_quads(
    quad_corners: np.ndarray,
    target_bbox: np.ndarray,
    out_idx_frac: np.ndarray,
    target_x: np.ndarray,
    target_y: np.ndarray,
    quad_ij0: np.ndarray,
) -> None:
    """Fill fractional source pixel indices for target pixels covered by source quads.

    For each source quadrilateral, this function iterates over all overlapping
    target pixels and computes barycentric coordinates using two triangle
    decompositions.
    """
    n_quads = quad_corners.shape[2]

    for q in range(n_quads):
        x0, y0, x1, y1 = target_bbox[:, q]

        c0 = quad_corners[:, 0, q]
        c1 = quad_corners[:, 1, q]
        c2 = quad_corners[:, 2, q]
        c3 = quad_corners[:, 3, q]

        # Triangle A
        dx1 = c1[0] - c0[0]
        dx2 = c2[0] - c0[0]
        dy1 = c1[1] - c0[1]
        dy2 = c2[1] - c0[1]
        det_a = dx1 * dy2 - dx2 * dy1

        # Triangle B
        dx1b = c2[0] - c3[0]
        dx2b = c1[0] - c3[0]
        dy1b = c2[1] - c3[1]
        dy2b = c1[1] - c3[1]
        det_b = dx1b * dy2b - dx2b * dy1b

        for i in range(x0, x1 + 1):
            tx = target_x[i]
            for j in range(y0, y1 + 1):
                if not np.isnan(out_idx_frac[0, j, i]):
                    continue

                ty = target_y[j]

                # Triangle A
                if det_a != 0.0:
                    fx = ((tx - c0[0]) * dy2 - (ty - c0[1]) * dx2) / det_a
                    fy = (dx1 * (ty - c0[1]) - dy1 * (tx - c0[0])) / det_a
                else:
                    fx = fy = np.nan
                if not np.isnan(fx) and (
                    fx >= -FX_MIN and fy >= -FY_MIN and fx + fy <= FXFY_MAX
                ):
                    out_idx_frac[0, j, i] = quad_ij0[0, q] + fx
                    out_idx_frac[1, j, i] = quad_ij0[1, q] + fy
                    continue

                # Triangle B
                if det_b != 0.0:
                    fx = ((tx - c3[0]) * dy2b - (ty - c3[1]) * dx2b) / det_b
                    fy = (dx1b * (ty - c3[1]) - dy1b * (tx - c3[0])) / det_b
                else:
                    fx = fy = np.nan
                if not np.isnan(fx) and (
                    fx >= -FX_MIN and fy >= -FY_MIN and fx + fy <= FXFY_MAX
                ):
                    out_idx_frac[0, j, i] = quad_ij0[0, q] + 1.0 - fx
                    out_idx_frac[1, j, i] = quad_ij0[1, q] + 1.0 - fy


def overlapping_bboxes(bbox, target_bboxes):
    block_i = []
    block_j = []
    for i in range(target_bboxes.shape[1]):
        for j in range(target_bboxes.shape[2]):
            if bbox_overlap(bbox, target_bboxes[:, i, j]) > 0:
                block_i.append(i)
                block_j.append(j)
    return block_i, block_j


def _xy_bbox_block(x_coords: np.ndarray, y_coords: np.ndarray):
    x_edges = np.concatenate([x_coords[:, 0], x_coords[:, -1]])
    y_edges = np.concatenate([y_coords[0, :], y_coords[-1, :]])
    bbox = np.array(
        [x_edges.min(), y_edges.min(), x_edges.max(), y_edges.max()],
        dtype=x_coords.dtype,
    )
    return bbox[:, None, None]


def _get_xy_bboxes(gm_2d: GridMapping):
    return da.map_blocks(
        _xy_bbox_block,
        gm_2d.x_coords.data,
        gm_2d.y_coords.data,
        dtype=gm_2d.x_coords.dtype,
        chunks=(4, 1, 1),
    )


def _compute_source_tile_indexing(
    source_gm: GridMapping, target_gm: GridMapping
) -> SourceTileIndexing:
    target_xy_bboxes = target_gm.xy_bboxes
    # add buffer to each tile
    target_xy_bboxes[:, 0] -= source_gm.x_res
    target_xy_bboxes[:, 2] += source_gm.x_res
    target_xy_bboxes[:, 1] -= source_gm.y_res
    target_xy_bboxes[:, 3] += source_gm.y_res

    source_xy_bboxes = _get_xy_bboxes(source_gm).compute()
    src_ij_bboxes = np.full_like(target_xy_bboxes, np.nan)

    tasks = []
    meta = []
    for tile_idx, bbox in enumerate(target_xy_bboxes):
        block_i, block_j = overlapping_bboxes(bbox, source_xy_bboxes)
        if not block_i:
            continue
        i_min = source_gm.tile_height * np.min(block_i)
        i_max = source_gm.tile_height * (np.max(block_i) + 1)
        j_min = source_gm.tile_width * np.min(block_j)
        j_max = source_gm.tile_width * (np.max(block_j) + 1)

        y_coords_sub = source_gm.y_coords[i_min:i_max, j_min:j_max].data
        x_coords_sub = source_gm.x_coords[i_min:i_max, j_min:j_max].data

        mask = (
            (x_coords_sub >= bbox[0])
            & (x_coords_sub <= bbox[2])
            & (y_coords_sub >= bbox[1])
            & (y_coords_sub <= bbox[3])
        )
        rows = da.any(mask, axis=1)
        cols = da.any(mask, axis=0)
        row_idxs = da.arange(rows.shape[0], chunks=rows.chunks)
        col_idxs = da.arange(cols.shape[0], chunks=cols.chunks)
        valid_rows = da.where(rows, row_idxs, np.nan)
        valid_cols = da.where(cols, col_idxs, np.nan)

        tasks.append(
            (
                da.nanmin(valid_rows),
                da.nanmax(valid_rows),
                da.nanmin(valid_cols),
                da.nanmax(valid_cols),
            )
        )
        meta.append((tile_idx, i_min, j_min))

    results = da.compute(*tasks)
    for (rmin, rmax, cmin, cmax), (tile_idx, i_min, j_min) in zip(results, meta):
        if np.isnan(rmin):
            continue
        src_ij_bboxes[tile_idx, 1] = rmin + i_min - 1
        src_ij_bboxes[tile_idx, 3] = rmax + i_min + 1
        src_ij_bboxes[tile_idx, 0] = cmin + j_min - 1
        src_ij_bboxes[tile_idx, 2] = cmax + j_min + 1
    target_block_j = int(np.ceil(target_gm.height / target_gm.tile_height))
    target_block_i = int(np.ceil(target_gm.width / target_gm.tile_width))
    src_ij_bboxes = src_ij_bboxes.reshape(
        (target_block_j, target_block_i, 4)
    ).transpose((2, 0, 1))

    # Extend bounding box indices to match the largest bounding box.
    # This ensures uniform chunk sizes, which are required for da.map_blocks.
    i_diff = src_ij_bboxes[2] - src_ij_bboxes[0]
    j_diff = src_ij_bboxes[3] - src_ij_bboxes[1]
    i_diff_max = np.nanmax(i_diff) + 1
    j_diff_max = np.nanmax(j_diff) + 1
    i_half = (i_diff_max - i_diff) // 2
    j_half = (j_diff_max - j_diff) // 2
    src_ij_bboxes[0] -= i_half
    src_ij_bboxes[2] = src_ij_bboxes[0] + i_diff_max
    src_ij_bboxes[1] -= j_half
    src_ij_bboxes[3] = src_ij_bboxes[1] + j_diff_max

    # assign padding if needed
    i_min = np.nanmin(src_ij_bboxes[0])
    i_max = np.nanmax(src_ij_bboxes[2])
    j_min = np.nanmin(src_ij_bboxes[[1, 3]])
    j_max = np.nanmax(src_ij_bboxes[[1, 3]])
    pad_width = (
        (-min(0, int(j_min)), max(0, int(j_max - source_gm.height))),
        (-min(0, int(i_min)), max(0, int(i_max - source_gm.width))),
    )
    src_ij_bboxes[[1, 3]] += pad_width[0][0]
    src_ij_bboxes[[0, 2]] += pad_width[1][0]

    src_ij_bboxes = np.where(np.isnan(src_ij_bboxes), -1, src_ij_bboxes).astype(
        np.int32
    )
    tile_size = (int(j_diff_max), int(i_diff_max))
    size = (
        int(j_diff_max * src_ij_bboxes.shape[1]),
        int(i_diff_max * src_ij_bboxes.shape[2]),
    )

    return SourceTileIndexing(
        ij_bboxes=src_ij_bboxes,
        pad_width=pad_width,
        output_size=size,
        tile_size=tile_size,
    )


def _rectify_data_array(
    data: da.Array,
    indexing: SourceTileIndexing,
    pixel_target_ij: da.Array,
    target_gm: GridMapping,
    interp_method: SpatialInterpMethod,
    fill_value: FloatInt,
) -> da.Array:

    expanded = data.ndim == 2
    if expanded:
        data = data[None, ...]

    is_numpy_array = False
    if isinstance(data, np.ndarray):
        is_numpy_array = True
        data = da.asarray(data)

    tiled = _reorganize_tiled_array(data, indexing, fill_value)

    rectified_chunks = []
    offset = 0

    for chunk in data.chunks[0]:
        rectified_chunks.append(
            da.map_blocks(
                _rectify_block,
                pixel_target_ij,
                tiled[offset : offset + chunk],
                interp_method=interp_method,
                fill_value=fill_value,
                dtype=data.dtype,
                chunks=(chunk, *pixel_target_ij.chunks[1:]),
            )
        )
        offset += chunk

    result = da.concatenate(rectified_chunks, axis=0)

    if expanded:
        result = result[0]

    if is_numpy_array and not target_gm.is_tiled:
        result = result.compute()

    return result


def _rectify_block(
    pixel_target_ij: np.ndarray,
    data_array: np.ndarray,
    interp_method: SpatialInterpMethodStr,
    fill_value: FloatInt,
) -> np.ndarray:
    ix, iy = pixel_target_ij

    out_shape = (data_array.shape[0],) + iy.shape
    data_rectified = np.full(out_shape, fill_value, dtype=data_array.dtype)

    valid = ~np.isnan(ix)
    if not np.any(valid):
        return data_rectified

    sampled = _sample_array_at_indices(data_array, ix[valid], iy[valid], interp_method)
    data_rectified[:, valid] = sampled
    return data_rectified
